import copy

import numpy as np
import scipy.fft as fft

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import NonSeparableWaveletTransformedImage
from mirp.settings.generic import SettingsClass
from mirp._imagefilters.generic import GenericFilter


class NonseparableWaveletFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = True
        self.ibsi_id = "LODD"

        # Set wavelet family
        self.wavelet_family: str | list[str] = settings.img_transform.nonseparable_wavelet_families

        # Wavelet decomposition level
        self.decomposition_level: int | list[int] = settings.img_transform.nonseparable_wavelet_decomposition_level

        # Riesz transformation settings.
        self.riesz_order: None | list[int] | list[list[int]] = None
        self.riesz_steered: bool = False
        self.riesz_sigma: None | float | list[float] = None
        if settings.img_transform.has_riesz_filter(x=name):
            self.riesz_order = settings.img_transform.riesz_order

            if settings.img_transform.has_steered_riesz_filter(x=name):
                self.riesz_steered = True
                self.riesz_sigma = settings.img_transform.riesz_filter_tensor_sigma

            # Riesz transformed filters are not IBSI-compliant
            self.ibsi_compliant = False

        # Set boundary condition
        self.mode = settings.img_transform.nonseparable_wavelet_boundary_condition

        # Set response.
        self.response = settings.img_transform.nonseparable_wavelet_response

    def generate_object(self):
        # Generator for transformation objects.
        wavelet_family = copy.deepcopy(self.wavelet_family)
        if not isinstance(wavelet_family, list):
            wavelet_family = [wavelet_family]

        decomposition_level = copy.deepcopy(self.decomposition_level)
        if not isinstance(decomposition_level, list):
            decomposition_level = [decomposition_level]

        riesz_order = copy.deepcopy(self.riesz_order)
        if riesz_order is None:
            riesz_order = [None]
        elif not all(isinstance(riesz_order_set, list) for riesz_order_set in riesz_order):
            riesz_order = [riesz_order]

        riesz_sigma = copy.deepcopy(self.riesz_sigma)
        if not isinstance(riesz_sigma, list):
            riesz_sigma = [riesz_sigma]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_family in wavelet_family:
            for current_riesz_order in riesz_order:
                for current_riesz_sigma in riesz_sigma:
                    for current_decomposition_level in decomposition_level:
                        filter_object = copy.deepcopy(self)
                        filter_object.wavelet_family = current_family
                        filter_object.decomposition_level = current_decomposition_level
                        filter_object.riesz_order = current_riesz_order
                        filter_object.riesz_sigma = current_riesz_sigma

                        yield filter_object

    def transform(self, image: GenericImage) -> NonSeparableWaveletTransformedImage:
        # Create placeholder non-separable wavelet response map.
        response_map = NonSeparableWaveletTransformedImage(
            image_data=None,
            wavelet_family=self.wavelet_family,
            decomposition_level=self.decomposition_level,
            response_type=self.response,
            boundary_condition=self.mode,
            riesz_order=self.riesz_order,
            riesz_steering=self.riesz_steered,
            riesz_sigma_parameter=self.riesz_sigma,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Create voxel grid
        response_voxel_grid = self.convolve(voxel_grid=image.get_voxel_grid())

        # Store the voxel grid in the ImageObject.
        response_map.set_voxel_grid(voxel_grid=response_voxel_grid)

        return response_map

    def shannon_filter(self, filter_size):
        """
        Set up the shannon filter in the Fourier domain.
        @param filter_size: Size of the filter. By default, equal to the size of the image.
        """

        # Get the distance grid.
        distance_grid, max_frequency = self.get_distance_grid(filter_size=filter_size)

        # Set up a wavelet filter for the decomposition specifically.
        wavelet_filter = np.zeros(distance_grid.shape, dtype=float)

        # Set the mask for the filter.
        mask = np.logical_and(distance_grid >= max_frequency / 2.0, distance_grid <= max_frequency)

        # Update the filter.
        wavelet_filter[mask] += 1.0

        return wavelet_filter

    def simoncelli_filter(self, filter_size):
        """
        Set up the simoncelli filter in the Fourier domain.
        @param filter_size: Size of the filter. By default, equal to the size of the image.
        """

        # Get the distance grid.
        distance_grid, max_frequency = self.get_distance_grid(filter_size=filter_size)

        # Set up a wavelet filter for the decomposition specifically.
        wavelet_filter = np.zeros(distance_grid.shape, dtype=float)

        # Set the mask for the filter.
        mask = np.logical_and(distance_grid >= max_frequency / 4.0,
                              distance_grid <= max_frequency)

        # Update the filter.
        wavelet_filter[mask] += np.cos(np.pi / 2.0 * np.log2(2.0 * distance_grid[mask] / max_frequency))

        return wavelet_filter

    def get_distance_grid(self, filter_size):
        """
        Create the distance grid.
        @param filter_size: Size of the filter. By default equal to the size of the image.
        """
        # Set up filter shape
        if filter_size is not None:
            filter_size = np.array(filter_size)
            if self.separate_slices:
                filter_shape = (filter_size[1], filter_size[2])
            else:
                filter_shape = (filter_size[0], filter_size[1], filter_size[2])
        else:
            if self.separate_slices:
                filter_shape = (filter_size, filter_size)

            else:
                filter_shape = (filter_size, filter_size, filter_size)

        # Determine the grid center.
        grid_center = (np.array(filter_shape, dtype=float) - 1.0) / 2.0

        # Determine distance from center
        distance_grid = list(np.indices(filter_shape, sparse=True))
        distance_grid = [(distance_grid[ii] - center_pos) / center_pos for ii, center_pos in enumerate(grid_center)]

        # Compute the distances in the grid.
        if len(filter_shape) == 2:
            distance_grid = np.sqrt(np.sum(
                np.power(np.meshgrid(distance_grid[0], distance_grid[1]), 2.0), axis=0)
            )
        elif len(filter_shape) == 3:
            distance_grid = np.sqrt(np.sum(
                np.power(np.meshgrid(distance_grid[0], distance_grid[1], distance_grid[2]), 2.0), axis=0)
            )
        else:
            raise ValueError("Filter shape should have 2 or 3 dimensions.")

        # Set the Nyquist frequency
        decomposed_max_frequency = 1.0 / 2.0 ** (self.decomposition_level - 1.0)

        return distance_grid, decomposed_max_frequency

    def convolve(self, voxel_grid):

        from mirp._imagefilters.utilities import FilterSet2D, FilterSet3D

        # Create the kernel.
        if self.wavelet_family == "simoncelli":
            wavelet_kernel_f = self.simoncelli_filter(filter_size=voxel_grid.shape)

        elif self.wavelet_family == "shannon":

            wavelet_kernel_f = self.shannon_filter(filter_size=voxel_grid.shape)

        else:
            raise ValueError(f"The specified wavelet family is not implemented: {self.wavelet_family}")

        if self.separate_slices:
            # Create filter set, and assign wavelet filter. Note the ifftshift that is present to go from a centric
            # to quadrant FFT representation.
            filter_set = FilterSet2D(filter_set=fft.ifftshift(wavelet_kernel_f),
                                     transformed=True,
                                     pad_image=False,
                                     riesz_order=self.riesz_order,
                                     riesz_steered=self.riesz_steered,
                                     riesz_sigma=self.riesz_sigma)

            # Create the response map.
            response_map = filter_set.convolve(voxel_grid=voxel_grid,
                                               mode=self.mode,
                                               response=self.response,
                                               axis=0)
        else:
            # Create filter set, and assign wavelet filter. Note the ifftshift that is present to go from a centric
            # to quadrant FFT representation.
            filter_set = FilterSet3D(filter_set=fft.ifftshift(wavelet_kernel_f),
                                     transformed=True,
                                     pad_image=False,
                                     riesz_order=self.riesz_order,
                                     riesz_steered=self.riesz_steered,
                                     riesz_sigma=self.riesz_sigma)

            # Create the response map.
            response_map = filter_set.convolve(voxel_grid=voxel_grid,
                                               mode=self.mode,
                                               response=self.response)

        return response_map



import copy

import numpy as np
import pandas as pd

from mirp._features.texture_matrix import Matrix


class MatrixNGTDM(Matrix):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Occurrence matrix
        self.pi: pd.DataFrame | None = None

        # Expanded occurrence matrix
        self.pij: pd.DataFrame | None = None

        # Number of intensity bins (grey levels)
        self.n_g: int | None = None

        # Number of intensity bins (grey levels) that are represented
        self.n_p: int | None = None

    def compute(
            self,
            data: pd.DataFrame | None = None,
            image_dimension: tuple[int, int, int] | None = None,
            **kwargs
    ):
        # Check if data actually exists
        if data is None:
            return

        # Check if the roi contains any masked voxels. If this is not the case, don't construct the GLRLM.
        if not np.any(data.roi_int_mask):
            return

        if image_dimension is None:
            raise ValueError("image_dimension cannot be None, but may not have been provided in the calling function.")

        # Create local copies of the image table
        if self.spatial_method == "3d":
            data = copy.deepcopy(data)

            directions = list(self._generate_neighbour_direction(
                d=1,
                metric="chebyshev",
                complete=True,
                dim3=True
            ))

        elif self.spatial_method in ["2d", "2.5d"]:
            data = copy.deepcopy(data[data.z == self.slice_id])
            data["index_id"] = np.arange(0, len(data))
            data["z"] = 0
            data = data.reset_index(drop=True)

            directions = list(self._generate_neighbour_direction(
                d=1,
                metric="chebyshev",
                complete=True,
                dim3=False
            ))

        else:
            directions = None
            self._spatial_method_error()

        # Set grey level of voxels outside ROI to NaN
        data.loc[data.roi_int_mask == False, "g"] = np.nan

        # Initialise sum of grey levels and number of neighbours.
        data["g_sum"] = 0
        data["n_nbrs"] = 0

        for direction in directions:
            # Determine potential valid neighbours.
            data["to_index"] = self.coord_to_index(
                x=data.x.values + direction[2],
                y=data.y.values + direction[1],
                z=data.z.values + direction[0],
                dims=image_dimension
            )

            # Get grey level value from transitions
            data["to_g"] = self._lookup_intensity(
                x=data.g.values,
                index=data.to_index.values
            )

            # Determine which voxels have valid neighbours
            valid_index = np.isfinite(data.to_g)

            # Sum grey level and increase neighbour counter
            data.loc[valid_index, "g_sum"] += data.loc[valid_index, "to_g"]
            data.loc[valid_index, "n_nbrs"] += 1

        # Calculate average neighbourhood grey level
        data["g_nbr_avg"] = data.g_sum / data.n_nbrs

        # Work with voxels without a missing grey value and with a valid neighbourhood
        data = data[np.logical_and(np.isfinite(data.g_nbr_avg), data.roi_int_mask)]

        # Determine contribution to s per voxel
        data["s_sub"] = np.abs(data.g - data.g_nbr_avg)

        # Drop superfluous columns
        data = data.drop(
            labels=["index_id", "x", "y", "z", "g_sum", "n_nbrs", "to_index", "to_g", "g_nbr_avg", "roi_int_mask"],
            axis=1
        )

        # Sum occurrences of grey level transitions after making the matrix symmetric.
        data = data.groupby(by="g")
        matrix = data.sum().join(pd.DataFrame(data.size(), columns=["n"])).reset_index()

        # Rename columns
        matrix.columns = ["i", "s", "n"]

        # Set the number of voxels
        self.n_voxels = np.sum(matrix.n)

        # Add matrix to object
        self.matrix = matrix

    def set_values_from_matrix(self, intensity_range: tuple[int, int], **kwargs):
        from mirp._features.utilities import rep

        if self.is_empty():
            return

        # Neighbourhood occurrence.
        self.pi = copy.deepcopy(self.matrix)
        self.pi["pi"] = self.pi.n / np.sum(self.pi.n)

        # Number of grey levels.
        self.n_g = int(intensity_range[1] - intensity_range[0] + 1)

        # Number of valid grey levels.
        self.n_p = len(self.pi)

        # Cartesian representation of neighbourhood occurrence.
        pi = copy.deepcopy(self.pi).drop(columns="n").rename(columns={"s": "si"})
        pj = copy.deepcopy(self.pi).drop(columns="n").rename(columns={"i": "j", "s": "sj", "pi": "pj"})
        pij = pd.DataFrame({
            "i": np.repeat(self.pi.i.values, self.n_p),
            "j": np.tile(self.pi.i.values, self.n_p)
        })
        pij = pd.merge(pij, pj, on="j")
        self.pij = pd.merge(pij, pi, on="i")

        # Neighbourhood occurrence with missing intensity levels.
        levels = np.arange(start=0, stop=self.n_g) + 1
        missing_levels = levels[np.logical_not(np.isin(levels, self.pi.i))]
        n_missing = len(missing_levels)
        if n_missing > 0:
            self.pi = pd.concat([
                self.pi,
                pd.DataFrame({
                    "i": missing_levels,
                    "s": np.zeros(n_missing),
                    "n": np.zeros(n_missing),
                    "pi": np.zeros(n_missing)
                })
            ],
                ignore_index=True
            )

    @staticmethod
    def _get_grouping_columns():
        return ["i"]



import copy
import warnings
import itk
import numpy as np
import pandas as pd

from mirp._data_import.generic_file import MaskFile
from mirp._data_import.itk_file import ImageITKFile
from mirp._data_import.generic_file_stack import ImageFileStack, MaskFileStack


class ImageITKFileStack(ImageFileStack):

    def __init__(
            self,
            image_file_objects: list[ImageITKFile],
            **kwargs
    ):
        super().__init__(image_file_objects, **kwargs)

    def complete(self, remove_metadata=False, force=False):
        """
        Fills out missing attributes in an image stack. Image parameters in ITK stacks are fully determined by the
        origin of all slices in the stack. This method then sorts the image file objects
        by origin, and uses their relative positions to determine slice spacing and the orientation vector. ITK-image
        stacks, e.g. stacks of NIfTI or NRRD files should be very very rare, since such formats are specifically
        designed to address some of the weakness of the DICOM standard slice-based format.
        :param remove_metadata: Whether metadata should be removed after completing information.
        :param force: Whether attributes are forced to update or not.
        :return: nothing, attributes are updated in place.
        """
        # Load metadata of every slice.
        self.load_metadata()

        self._complete_modality()
        self._complete_sample_name()

        # Placeholders for slice positions.
        image_position_z = [0.0] * len(self.image_file_objects)
        image_position_y = [0.0] * len(self.image_file_objects)
        image_position_x = [0.0] * len(self.image_file_objects)

        for ii, image_object in enumerate(self.image_file_objects):
            slice_origin = np.array(image_object.image_metadata.GetOrigin())[::-1]

            image_position_z[ii] = slice_origin[0]
            image_position_y[ii] = slice_origin[1]
            image_position_x[ii] = slice_origin[2]

        # Order ascending position (DICOM: z increases from feet to head)
        position_table = pd.DataFrame({
            "original_object_order": list(range(len(self.image_file_objects))),
            "position_z": image_position_z,
            "position_y": image_position_y,
            "position_x": image_position_x
        }).sort_values(by=["position_z", "position_y", "position_x"])

        # Set image spacing. Compute the distance between the origins of the slices. This is the slice spacing.
        image_slice_spacing = np.sqrt(
            np.power(np.diff(position_table.position_x.values), 2.0) +
            np.power(np.diff(position_table.position_y.values), 2.0) +
            np.power(np.diff(position_table.position_z.values), 2.0))

        # Find the smallest slice spacing.
        min_slice_spacing = np.min(image_slice_spacing)
        if min_slice_spacing == 0.0:
            warnings.warn(
                "Images files contain overlapping origins. Attempting to sort image files by numeric name "
                f"patterns. [{self.describe_self()}]",
                UserWarning
            )
            self.sort_image_objects_by_file()

            image_object = copy.deepcopy(self.image_file_objects[0])
            image_object.complete()

            if self.image_origin is None:
                self.image_origin = image_object.image_origin

            if self.image_spacing is None:
                self.image_spacing = image_object.image_spacing

            if self.image_orientation is None:
                self.image_orientation = image_object.image_orientation

            if self.image_dimension is None:
                self.image_dimension = tuple([
                    len(self.image_file_objects), image_object.image_dimension[1], image_object.image_dimension[2]
                ])

        else:
            # Sort image file objects.
            self.image_file_objects = [
                self.image_file_objects[position_table.original_object_order[ii]]
                for ii in range(len(position_table))
            ]

        # Set image origin.
        if self.image_origin is None:
            self.image_origin = tuple(np.array(self.image_file_objects[0].image_metadata.GetOrigin())[::-1])

        # Find how much other slices differ.
        image_slice_spacing_multiplier = image_slice_spacing / min_slice_spacing

        if np.any(image_slice_spacing_multiplier > 1.2):
            warnings.warn(
                f"Inconsistent distance between slice origins of subsequent slices: {np.unique(image_slice_spacing)}. "
                f"Slices cannot be aligned correctly. This is likely due to missing slices. "
                f"MIRP will attempt to interpolate the missing slices and their ROI masks for volumetric analysis. "
                f"[{self.describe_self()}]",
                UserWarning
            )

            # Update slice positions.
            self.slice_positions = list(np.cumsum(np.insert(np.around(image_slice_spacing, 5), 0, 0.0)))

        # Determine image slice spacing.
        image_slice_spacing = np.around(np.mean(image_slice_spacing[image_slice_spacing_multiplier <= 1.2]), 5)

        # Set image spacing.
        image_spacing = np.array(self.image_file_objects[0].image_metadata.GetSpacing())[::-1]
        if self.image_spacing is None:
            if image_spacing[0] == image_slice_spacing:
                self.image_spacing = tuple(image_spacing)
            else:
                self.image_spacing = tuple([image_slice_spacing, image_spacing[1], image_spacing[2]])

        # Read orientation metadata.
        image_orientation = np.reshape(np.ravel(itk.array_from_matrix(
            self.image_file_objects[0].image_metadata.GetDirection()))[::-1], [3, 3])

        # Add (Zx, Zy, Zz)
        z_orientation = np.array([
            np.around(np.min(np.diff(position_table.position_z.values)), 5),
            np.around(np.min(np.diff(position_table.position_y.values)), 5),
            np.around(np.min(np.diff(position_table.position_x.values)), 5)
        ]) / image_slice_spacing

        # Replace z-orientation and set image_orientation.
        image_orientation[0, :] = z_orientation
        if self.image_orientation is None:
            self.image_orientation = image_orientation

        # Set dimension
        image_dimension = np.array(self.image_metadata.GetSize())[::-1]

        if self.image_dimension:
            self.image_dimension = tuple([len(position_table), image_dimension[1], image_dimension[2]])

        # Check if the complete data passes verification.
        self.check(raise_error=True, remove_metadata=False)


class MaskITKFileStack(ImageITKFileStack, MaskFileStack):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def complete(self, remove_metadata=False, force=False):

        super().complete(remove_metadata=False, force=force)

        image_object: MaskFile = copy.deepcopy(self.image_file_objects[0])
        image_object.complete(remove_metadata=False)

        self.roi_name = image_object.roi_name



from mirp.deep_learning_preprocessing import deep_learning_preprocessing, deep_learning_preprocessing_generator
from mirp.extract_features_and_images import extract_features, extract_features_generator, extract_images, \
    extract_images_generator, extract_features_and_images, extract_features_and_images_generator
from mirp.extract_image_parameters import extract_image_parameters
from mirp.extract_mask_labels import extract_mask_labels
from mirp.utilities.config_utilities import get_data_xml, get_settings_xml

__all__ = [
    "deep_learning_preprocessing",
    "deep_learning_preprocessing_generator",
    "extract_features",
    "extract_features_generator",
    "extract_images",
    "extract_images_generator",
    "extract_features_and_images",
    "extract_features_and_images_generator",
    "extract_image_parameters",
    "extract_mask_labels",
    "get_data_xml",
    "get_settings_xml"
]



import numpy as np

from mirp._images.generic_image import GenericImage


class MRImage(GenericImage):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def bias_field_correction(
            self,
            n_fitting_levels: int = 3,
            n_max_iterations: None | int | list[int] = None,
            convergence_threshold: float = 0.001,
            mask: None | np.ndarray = None,
            in_place: bool = True
    ) -> None | GenericImage:

        import itk

        if self.is_empty():
            return

        if mask is None:
            mask = np.ones(self.image_dimension, dtype=np.uint8)

        if n_max_iterations is None:
            n_max_iterations = [50 for ii in range(n_fitting_levels)]

        # Create ITK input masks
        input_image = itk.GetImageFromArray(self.get_voxel_grid())
        input_image.SetSpacing(self.image_spacing[::-1])
        input_mask = itk.GetImageFromArray(mask.astype(np.uint8))
        input_mask.SetSpacing(self.image_spacing[::-1])

        # Set number of threads
        threader = itk.MultiThreaderBase.New()
        threader.SetGlobalDefaultNumberOfThreads(1)

        # Start N4 bias correction
        corrector = itk.N4BiasFieldCorrectionImageFilter[type(input_image), type(input_mask), type(input_image)].New(input_image, input_mask)
        corrector.SetMaximumNumberOfIterations(n_max_iterations)
        corrector.SetNumberOfFittingLevels(n_fitting_levels)
        corrector.SetConvergenceThreshold(convergence_threshold)
        corrector.Update()
        output_image = corrector.GetOutput()

        # Save bias-corrected image.
        if not in_place:
            image = self.copy(drop_image=True)
            image.set_voxel_grid(voxel_grid=itk.GetArrayFromImage(output_image).astype(float))
            return image
        else:
            self.set_voxel_grid(voxel_grid=itk.GetArrayFromImage(output_image).astype(float))



import os.path
import numpy as np

from mirp._data_import.generic_file import ImageFile, MaskFile


class ImageNumpyFile(ImageFile):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        """
        Is the image potentially stackable?
        :param stack_images: One of auto, yes or no. By default (auto), images are stackable if an image object
        represents a single slice.
        :return: boolean value.
        """
        if stack_images in ["auto", "yes"]:
            if self.image_dimension is None:
                raise ValueError(
                    f"The image_dimension argument is expected to be set. Call load_metadata to set this attribute. "
                    f"[{self.describe_self()}]"
                )

            if len(self.image_dimension) < 3:
                return True
            elif self.image_dimension[0] == 1:
                return True
            else:
                return False

        elif stack_images == "no":
            return False
        else:
            raise ValueError(
                f"The stack_images argument is expected to be one of yes, auto, or no. Found: {stack_images}. "
                f"[{self.describe_self()}]"
            )

    def check(self, raise_error=False, remove_metadata=True) -> bool:

        # Perform general checks.
        if not super().check(raise_error=raise_error):
            return False

        # Read Numpy file.
        image_data = np.load(file=self.file_path, mmap_mode="r")

        # Check that the contents are in fact a ndarray.
        if not isinstance(image_data, np.ndarray):
            if raise_error:
                raise TypeError(
                    f"The current numpy dataset {self.file_path} does not have the expected content. "
                    f"Found: {type(image_data)}. Expected: numpy.ndarray. [{self.describe_self()}]"
                )

            return False

        # Check dimensions.
        if not 0 < image_data.ndim <= 3:
            if raise_error:
                raise ValueError(
                    f"The current numpy dataset has as an unexpected number of dimensions: "
                    f"Found: {image_data.ndim}. Expected: 1, 2, or 3 dimensions. [{self.describe_self()}]"
                )

            return False

        return True

    def _complete_image_origin(self, force=False, frame_id=None):
        if self.image_origin is None:
            self.load_metadata()
            n_dim = len(self.image_metadata.shape)
            self.image_origin = tuple([0.0] * n_dim)

    def _complete_image_orientation(self, force=False, frame_id=None):
        if self.image_orientation is None:
            self.load_metadata()
            n_dim = len(self.image_metadata.shape)
            self.image_orientation = np.identity(n_dim, dtype=float)

    def _complete_image_spacing(self, force=False, frame_id=None):
        if self.image_spacing is None:
            self.load_metadata()
            n_dim = len(self.image_metadata.shape)
            self.image_spacing = tuple([1.0] * n_dim)

    def _complete_image_dimensions(self, force=False):
        if self.image_dimension is None:
            self.load_metadata()
            self.image_dimension = tuple(self.image_metadata.shape)

    def create(self):
        return self

    def load_metadata(self, **kwargs):
        if self.image_metadata is not None:
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. "
                f"[{self.describe_self()}]"
            )

        # `Lazy load the data
        self.image_metadata = np.load(file=self.file_path, mmap_mode="r")

    def load_data(self, **kwargs):
        if self.image_data is not None:
            self.update_image_data()
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. "
                f"[{self.describe_self()}]"
            )

        self.image_data = np.load(self.file_path).astype(np.float32)
        self.update_image_data()


class MaskNumpyFile(ImageNumpyFile, MaskFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def load_data(self, **kwargs):
        if self.image_data is not None:
            self.update_image_data()
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The mask file could not be found at the expected location: {self.file_path}. [{self.describe_self()}]"
            )

        self.image_data = np.load(self.file_path).astype(int)
        self.update_image_data()



from functools import lru_cache
from typing import Generator

import numpy as np
import pandas as pd

from mirp._features.cm_matrix import MatrixCM
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters
from mirp._features.texture_features import FeatureTexture
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class FeatureCM(FeatureTexture):

    def __init__(
            self,
            distance: int,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Lookup distance (in voxels)
        self.distance = int(distance)

        # Features are always computed from symmetric co-occurrence matrices.
        self.symmetric_matrix = True

        # Perform close crop for CM.
        self.cropping_distance = 0.0

    def get_matrix(
            self,
            image: GenericImage,
            mask: BaseMask,
    ) -> list[MatrixCM]:
        # First discretise the image.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Then get matrix or matrices
        matrix_list = self._get_matrix(
            image=image,
            mask=mask,
            spatial_method=self.spatial_method,
            distance=self.distance
        )

        return matrix_list

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_matrix(
            image: GenericImage,
            mask: BaseMask,
            spatial_method: str,
            distance: int
    ) -> list[MatrixCM]:
        # Represent image and mask as a dataframe.
        data = mask.as_pandas_dataframe(
            image=image,
            intensity_mask=True
        )

        # Instantiate a helper copy of the current class to be able to use class methods without tying the cache to the
        # instance of the original object from which this method is called.
        matrix_instance = MatrixCM(
            spatial_method=spatial_method,
            distance=distance
        )

        # Compute the required matrices.
        matrix_list = list(matrix_instance.generate(prototype=MatrixCM, n_slices=image.image_dimension[0], distance=distance))
        for matrix in matrix_list:
            matrix.compute(data=data, image_dimension=image.image_dimension)

        # Merge according to the spatial method.
        matrix_list = matrix_instance.merge(matrix_list, prototype=MatrixCM, distance=distance)

        # Compute additional values from the individual matrices.
        for matrix in matrix_list:
            matrix.set_values_from_matrix(intensity_range=mask.intensity_range)

        return matrix_list

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureCM):
            self._get_matrix.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_matrix.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Compute or retrieve matrices from cache.
        matrices = self.get_matrix(image=image, mask=mask)

        # Compute feature value from matrices, and average over matrices.
        values = [self._compute(matrix=matrix) for matrix in matrices]
        if np.all(np.isnan(values)):
            self.value = np.nan
        else:
            self.value = np.nanmean(values)

    @staticmethod
    def _compute(matrix: MatrixCM):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + ["d" + str(self.distance)]
                + self._get_spatial_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureCMJointMax(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - joint maximum"
        self.abbr_name = "cm_joint_max"
        self.ibsi_id = "GYBY"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.max(matrix.pij.pij)


class FeatureCMJointAverage(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - joint average"
        self.abbr_name = "cm_joint_avg"
        self.ibsi_id = "60VM"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return matrix.mu


class FeatureCMJointVariance(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - joint variance"
        self.abbr_name = "cm_joint_var"
        self.ibsi_id = "UR99"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum((matrix.pij.i - matrix.mu) ** 2.0 * matrix.pij.pij)


class FeatureCMJointEntropy(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - joint entropy"
        self.abbr_name = "cm_joint_entr"
        self.ibsi_id = "TU9B"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return -np.sum(matrix.pij.pij * np.log2(matrix.pij.pij))


class FeatureCMDifferenceAverage(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - difference average"
        self.abbr_name = "cm_diff_avg"
        self.ibsi_id = "TF7R"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pimj.k * matrix.pimj.pimj)


class FeatureCMDifferenceVariance(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - difference variance"
        self.abbr_name = "cm_diff_var"
        self.ibsi_id = "D3YU"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.pimj.k * matrix.pimj.pimj)
        return np.sum((matrix.pimj.k - mu) ** 2.0 * matrix.pimj.pimj)


class FeatureCMDifferenceEntropy(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - difference entropy"
        self.abbr_name = "cm_diff_entr"
        self.ibsi_id = "NTRS"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return -np.sum(matrix.pimj.pimj * np.log2(matrix.pimj.pimj))


class FeatureCMSumAverage(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - sum average"
        self.abbr_name = "cm_sum_avg"
        self.ibsi_id = "ZGXS"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pipj.k * matrix.pipj.pipj)


class FeatureCMSumVariance(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - sum variance"
        self.abbr_name = "cm_sum_var"
        self.ibsi_id = "OEEB"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.pipj.k * matrix.pipj.pipj)
        return np.sum((matrix.pipj.k - mu) ** 2.0 * matrix.pipj.pipj)


class FeatureCMSumEntropy(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - sum entropy"
        self.abbr_name = "cm_sum_entr"
        self.ibsi_id = "P6QZ1"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return -np.sum(matrix.pipj.pipj * np.log2(matrix.pipj.pipj))


class FeatureCMAngularSecondMoment(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - angular second moment"
        self.abbr_name = "cm_energy"
        self.ibsi_id = "8ZQL"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pij.pij ** 2.0)


class FeatureCMContrast(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - contrast"
        self.abbr_name = "cm_contrast"
        self.ibsi_id = "ACUI"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum((matrix.pij.i - matrix.pij.j) ** 2.0 * matrix.pij.pij)


class FeatureCMDissimilarity(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - dissimilarity"
        self.abbr_name = "cm_dissimilarity"
        self.ibsi_id = "8S9J"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(np.abs(matrix.pij.i - matrix.pij.j) * matrix.pij.pij)


class FeatureCMInverseDifference(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - inverse difference"
        self.abbr_name = "cm_inv_diff"
        self.ibsi_id = "IB1Z"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pij.pij / (1.0 + np.abs(matrix.pij.i - matrix.pij.j)))


class FeatureCMNormalisedInverseDifference(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - normalised inverse difference"
        self.abbr_name = "cm_inv_diff_norm"
        self.ibsi_id = "NDRX"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pij.pij / (1.0 + np.abs(matrix.pij.i - matrix.pij.j) / matrix.n_g))


class FeatureCMInverseDifferenceMoment(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - inverse difference moment"
        self.abbr_name = "cm_inv_diff_mom"
        self.ibsi_id = "WF0Z"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pij.pij / (1.0 + (matrix.pij.i - matrix.pij.j) ** 2.0))


class FeatureCMNormalisedInverseDifferenceMoment(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - normalised inverse difference moment"
        self.abbr_name = "cm_inv_diff_mom_norm"
        self.ibsi_id = "1QCO"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pij.pij / (1.0 + (matrix.pij.i - matrix.pij.j) ** 2.0 / matrix.n_g ** 2.0))


class FeatureCMInverseVariance(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - inverse variance"
        self.abbr_name = "cm_inv_var"
        self.ibsi_id = "E8JP"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan

        # Select only off-diagonal elements.
        matrix = matrix.pij[matrix.pij.i != matrix.pij.j]
        if len(matrix) == 0:
            return np.nan

        return np.sum(matrix.pij / (matrix.i - matrix.j) ** 2.0)


class FeatureCMCorrelation(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - correlation"
        self.abbr_name = "cm_corr"
        self.ibsi_id = "NI2N"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan

        var_marg = np.sum((matrix.pi.i - matrix.mu_marg) ** 2.0 * matrix.pi.pi)

        if var_marg == 0.0:
            return 1.0
        else:
            return 1.0 / var_marg * (np.sum(matrix.pij.i * matrix.pij.j * matrix.pij.pij) - matrix.mu_marg ** 2.0)


class FeatureCMAutoCorrelation(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - autocorrelation"
        self.abbr_name = "cm_auto_corr"
        self.ibsi_id = "QWB0"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.pij.i * matrix.pij.j * matrix.pij.pij)


class FeatureCMClusterTendency(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - cluster tendency"
        self.abbr_name = "cm_clust_tend"
        self.ibsi_id = "DG8W"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum((matrix.pij.i + matrix.pij.j - 2.0 * matrix.mu_marg) ** 2.0 * matrix.pij.pij)


class FeatureCMClusterShade(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - cluster shade"
        self.abbr_name = "cm_clust_shade"
        self.ibsi_id = "7NFM"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum((matrix.pij.i + matrix.pij.j - 2.0 * matrix.mu_marg) ** 3.0 * matrix.pij.pij)


class FeatureCMClusterProminence(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - cluster prominence"
        self.abbr_name = "cm_clust_prom"
        self.ibsi_id = "AE86"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum((matrix.pij.i + matrix.pij.j - 2.0 * matrix.mu_marg) ** 4.0 * matrix.pij.pij)


class FeatureCMInformationCorrelation1(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - first measure of information correlation"
        self.abbr_name = "cm_info_corr1"
        self.ibsi_id = "R8DG"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan
        hxy = - np.sum(matrix.pij.pij * np.log2(matrix.pij.pij))
        hxy_1 = - np.sum(matrix.pij.pij * np.log2(matrix.pij.pi * matrix.pij.pj))
        hx = - np.sum(matrix.pi.pi * np.log2(matrix.pi.pi))
        if len(matrix.pij) == 1 or hx == 0.0:
            return 1.0
        else:
            return (hxy - hxy_1) / hx


class FeatureCMInformationCorrelation2(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - second measure of information correlation"
        self.abbr_name = "cm_info_corr2"
        self.ibsi_id = "JN9H"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan

        hxy = - np.sum(matrix.pij.pij * np.log2(matrix.pij.pij))
        hxy_2 = - np.sum(
            np.tile(matrix.pi.pi, len(matrix.pj))
            * np.repeat(matrix.pj.pj, len(matrix.pi))
            * np.log2(
                np.tile(matrix.pi.pi, len(matrix.pj))
                * np.repeat(matrix.pj.pj, len(matrix.pi))
            )
        )

        if hxy_2 < hxy:
            return 0.0
        else:
            return np.sqrt(1.0 - np.exp(-2.0 * (hxy_2 - hxy)))


class FeatureCMMaximumCorrelationCoefficient(FeatureCM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CM - maximum correlation coefficient"
        self.abbr_name = "cm_mcc"

        # MCC is not part of the IBSI corpus and has no assigned identifier.
        self.ibsi_compliant = False

    @staticmethod
    def _compute(matrix: MatrixCM) -> float:
        if matrix.is_empty():
            return np.nan

        # MCC is defined as the square root of the second largest eigen value of Q (Haralick, R.M., Shanmugam,
        # K. and Dinstein, I. (1973) ‘Textural Features for Image Classification’, IEEE transactions on systems, man,
        # and cybernetics, SMC-3(6), pp. 610–621. Available at: https://doi.org/10.1109/TSMC.1973.4309314.)

        # Special case: matrix with a single element.
        if matrix.n_g == 1.0:
            return 1.0

        # Initialise matrix for q.
        q_mat = np.zeros((int(matrix.n_g), int(matrix.n_g)), dtype=float)

        # Initialise full matrix for pij, pi and pj and fill. For pi and pj use 1s as default value to prevent division
        # by 0.
        pij = np.zeros((int(matrix.n_g), int(matrix.n_g)), dtype=float)
        pij[(matrix.pij.i.values.astype(int) - 1, matrix.pij.j.values.astype(int) - 1)] = matrix.pij.pij.values
        pi = np.ones((int(matrix.n_g)), dtype=float)
        pi[(matrix.pi.i.values.astype(int) - 1)] = matrix.pi.pi.values
        pj = np.ones((int(matrix.n_g)), dtype=float)
        pj[(matrix.pj.j.values.astype(int) - 1)] = matrix.pj.pj.values

        non_zero_elems = np.nonzero(pij)
        for kk in np.arange(len(non_zero_elems[0])):
            i = non_zero_elems[0][kk]
            j = non_zero_elems[1][kk]

            # sum of p(i,k) * p(j,k) / (px(i), py(k))
            q_mat[i,j] = np.sum(np.divide(np.multiply(pij[i, :], pij[j, :]), pi[i] * pj))

        # Compute eigen values and sort.
        eigen_values = np.linalg.eigvals(q_mat)
        eigen_values.sort()

        # Select second largest eigenvalue, and set minimum value to 0.
        second_largest_eigen_value = np.max([0.0, np.real(eigen_values[-2])])

        # Return square root of second largest eigenvalue.
        return np.real(np.sqrt(second_largest_eigen_value))


def get_cm_class_dict() -> dict[str, FeatureCM]:
    class_dict = {
        "cm_joint_max": FeatureCMJointMax,
        "cm_joint_avg": FeatureCMJointAverage,
        "cm_joint_var": FeatureCMJointVariance,
        "cm_joint_entr": FeatureCMJointEntropy,
        "cm_diff_avg": FeatureCMDifferenceAverage,
        "cm_diff_var": FeatureCMDifferenceVariance,
        "cm_diff_entr": FeatureCMDifferenceEntropy,
        "cm_sum_avg": FeatureCMSumAverage,
        "cm_sum_var": FeatureCMSumVariance,
        "cm_sum_entr": FeatureCMSumEntropy,
        "cm_energy": FeatureCMAngularSecondMoment,
        "cm_contrast": FeatureCMContrast,
        "cm_dissimilarity": FeatureCMDissimilarity,
        "cm_inv_diff": FeatureCMInverseDifference,
        "cm_inv_diff_norm": FeatureCMNormalisedInverseDifference,
        "cm_inv_diff_mom": FeatureCMInverseDifferenceMoment,
        "cm_inv_diff_mom_norm": FeatureCMNormalisedInverseDifferenceMoment,
        "cm_inv_var": FeatureCMInverseVariance,
        "cm_corr": FeatureCMCorrelation,
        "cm_auto_corr": FeatureCMAutoCorrelation,
        "cm_clust_tend": FeatureCMClusterTendency,
        "cm_clust_shade": FeatureCMClusterShade,
        "cm_clust_prom": FeatureCMClusterProminence,
        "cm_info_corr1": FeatureCMInformationCorrelation1,
        "cm_info_corr2": FeatureCMInformationCorrelation2,
        "cm_mcc": FeatureCMMaximumCorrelationCoefficient
    }

    return class_dict


def generate_cm_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureCM, None, None]:
    class_dict = get_cm_class_dict()
    cm_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_glcm_family():
        features = cm_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only CM-features, and return if none are present.
    features = [feature for feature in features if feature in cm_features]
    if len(features) == 0:
        return

    # Features are parametrised by the choice of discretisation parameters, spatial methods and distances.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for spatial_method in settings.glcm_spatial_method:
            for distance in settings.glcm_distance:
                for feature in features:
                    yield class_dict[feature](
                        spatial_method=spatial_method,
                        distance=distance,
                        **discretisation_parameters
                    )



from typing import Any
from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.dicom_multi_frame import ImageDicomMultiFrame
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomFileMR(ImageDicomFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        return True

    def create(self):
        return self

    def load_data(self, **kwargs):
        self.image_data = self.load_data_generic()

    def export_metadata(self, self_only=False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Receive coil name
        receive_coil = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1250),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9042)
        )
        if receive_coil is not None:
            dcm_meta_data += [("receive_coil_name", receive_coil)]

        # Receive coil manufacturer
        receive_coil_manufacturer = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9041),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9042)
        )
        if receive_coil_manufacturer is not None:
            dcm_meta_data += [("receive_coil_manufacturer", receive_coil_manufacturer)]

        # Receive coil type
        receive_coil_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9043),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9042)
        )
        if receive_coil_type is not None:
            dcm_meta_data += [("receive_coil_type", receive_coil_type)]

        # Transmit coil name
        transmit_coil = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1251),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9049)
        )
        if transmit_coil is not None:
            dcm_meta_data += [("transmit_coil_name", transmit_coil)]

        # Transmit coil manufacturer
        transmit_coil_manufacturer = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9050),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9049)
        )
        if transmit_coil_manufacturer is not None:
            dcm_meta_data += [("transmit_coil_manufacturer", transmit_coil_manufacturer)]

        # Transmit coil type
        transmit_coil_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9051),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9049)
        )
        if transmit_coil_type is not None:
            dcm_meta_data += [("transmit_coil_type", transmit_coil_type)]

        # Magnetic field strength
        magnetic_field_strength = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0087),
            tag_type="float"
        )
        if magnetic_field_strength is not None:
            dcm_meta_data += [("magnetic_field_strength", magnetic_field_strength)]

        # Scanning sequence
        scanning_sequence = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0020),
            tag_type="str"
        )
        if scanning_sequence is not None:
            dcm_meta_data += [("scanning_sequence", scanning_sequence)]

        # Scanning sequence variant
        scanning_sequence_variant = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0021),
            tag_type="str"
        )
        if scanning_sequence_variant is not None:
            dcm_meta_data += [("scanning_sequence_variant", scanning_sequence_variant)]

        # Sequence name
        scanning_sequence_name = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0024),
            tag_type="str"
        )
        if scanning_sequence_name is not None:
            dcm_meta_data += [("scanning_sequence_name", scanning_sequence_name)]

        # Acquisition contrast
        acquisition_contrast = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9209),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9226)
        )
        if acquisition_contrast is not None:
            dcm_meta_data += [("acquisition_contrast", acquisition_contrast)]

        # Scan options
        scan_options = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0022),
            tag_type="str"
        )
        if scan_options is not None:
            dcm_meta_data += [("scan_options", scan_options)]

        # Acquisition type
        acquisition_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0023),
            tag_type="str"
        )
        if acquisition_type is not None:
            dcm_meta_data += [("acquisition_type", acquisition_type)]

        # Parallel acquisition technique
        parallel_acquisition_technique = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9078),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9115)
        )
        if parallel_acquisition_technique is not None:
            dcm_meta_data += [("parallel_acquisition_technique", parallel_acquisition_technique)]

        # Repetition time
        repetition_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0080),
            tag_type="float",
            macro_dcm_seq=(0x0018, 0x9112)
        )
        if repetition_time is not None:
            dcm_meta_data += [("repetition_time", repetition_time)]

        # Echo time
        echo_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0081),
            tag_type="float",
            macro_dcm_seq=(0x0018, 0x9112)
        )
        if echo_time is not None:
            dcm_meta_data += [("echo_time", echo_time)]

        # Echo train length
        echo_train_length = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0091),
            tag_type="float",
            macro_dcm_seq=(0x0018, 0x9112)
        )
        if echo_train_length is not None:
            dcm_meta_data += [("echo_train_length", echo_train_length)]

        # Inversion time
        inversion_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0082),
            tag_type="float"
        )
        if inversion_time is not None:
            dcm_meta_data += [("inversion_time", inversion_time)]

        # Trigger time
        trigger_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1060),
            tag_type="float"
        )
        if trigger_time is not None:
            dcm_meta_data += [("trigger_time", trigger_time)]

        # Contrast/bolus agent
        contrast_agent = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0010),
            tag_type="str"
        )
        if contrast_agent is not None:
            dcm_meta_data += [("contrast_agent", contrast_agent)]

        # Contrast/bolus agent phase
        contrast_agent_phase = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9344),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9341)
        )
        if contrast_agent_phase is not None:
            dcm_meta_data += [("contrast_agent_phase", contrast_agent_phase)]

        metadata.update(dict(dcm_meta_data))
        return metadata


class ImageDicomFileMRMultiFrame(ImageDicomMultiFrame, ImageDicomFileMR):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import warnings
from typing import Any

import numpy as np

from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask


def discretise_image(
        image: GenericImage,
        mask: None | BaseMask,
        discretisation_method: None | str = "none",
        intensity_range: None | tuple[Any, Any] = None,
        bin_width: None | int = None,
        bin_number: None | int = None,
        in_place: bool = False
):
    if image.is_empty():
        return None, None

    if not in_place:
        image = image.copy()
        mask = mask.copy()

    if mask is None:
        mask_data = np.ones(image.image_dimension, dtype=bool)
    elif mask.roi_intensity is None or mask.roi_intensity.is_empty() or mask.roi_intensity.is_empty_mask():
        return None, None
    else:
        mask_data = mask.roi_intensity.get_voxel_grid()
        intensity_range = mask.intensity_range

    if discretisation_method is None or discretisation_method == "none":
        levels = np.unique(image.get_voxel_grid()[mask_data])

        # Check if voxels are discretised.
        if not np.all(np.fmod(levels, 1.0) == 0.0):
            raise ValueError(f"The 'none' transformation method can only be used for data that resemble bins.")

        if not np.min(levels) >= 1.0:
            raise ValueError(
                f"The 'none' transformation method requires integer (i.e. 1.0, 2.0, etc.) values with a minimum value "
                f"of 1."
            )

        discretisation_method = "none"
        discretised_voxels = image.get_voxel_grid()
        bin_number = np.max(levels)

    elif discretisation_method == "fixed_bin_number":
        min_intensity = np.min(image.get_voxel_grid()[mask_data])
        max_intensity = np.max(image.get_voxel_grid()[mask_data])

        if min_intensity < max_intensity:
            discretised_voxels = np.floor(
                bin_number * 1.0 * (image.get_voxel_grid() - min_intensity) / (max_intensity - min_intensity)) + 1.0
        else:
            discretised_voxels = np.zeros(image.image_dimension, dtype=float) + np.ceil(bin_number / 2.0)

        discretised_voxels[discretised_voxels < 1.0] = 1.0
        discretised_voxels[discretised_voxels >= bin_number * 1.0] = bin_number * 1.0

        # Set number of bins.
        image.discretisation_bin_number = bin_number

    elif discretisation_method == "fixed_bin_size":
        if intensity_range is None or np.isnan(intensity_range[0]):
            min_intensity = image.get_default_lowest_intensity()

            if min_intensity is None:
                raise ValueError(
                    "Discretisation using the Fixed Bin Size method requires that the lower bound of the "
                    "intensity range is set."
                )
            else:
                warnings.warn(
                    f"No lower bound of the intensity range was set for discretisation using the Fixed Bin "
                    f"Size method. A default value {min_intensity} was used."
                )
        else:
            min_intensity = intensity_range[0]

        # Bin voxels.
        discretised_voxels = np.floor((image.get_voxel_grid() - min_intensity) / (bin_width * 1.0)) + 1.0

        # Set voxels with grey level lower than 0.0 to 1.0. This may occur with non-roi voxels and voxels with the
        # minimum intensity.
        discretised_voxels[discretised_voxels <= 1.0] = 1.0

        # Determine number of bins
        bin_number = np.max(discretised_voxels[mask_data])

        # Limit to maximum number of bins
        discretised_voxels[discretised_voxels >= bin_number * 1.0] = bin_number * 1.0

        # Set bin width.
        image.discretisation_bin_width = bin_width

    elif discretisation_method == "fixed_bin_size_pyradiomics":
        # PyRadiomics (up to at least version 3.1.0) used a version of fixed bin size that is not IBSI-compliant.
        min_intensity = np.min(image.get_voxel_grid()[mask_data])
        min_intensity -= min_intensity % bin_width

        # Bin voxels.
        discretised_voxels = np.floor((image.get_voxel_grid() - min_intensity) / (bin_width * 1.0)) + 1.0

        # Set voxels with grey level lower than 0.0 to 1.0. This may occur with non-roi voxels and voxels with the
        # minimum intensity.
        discretised_voxels[discretised_voxels <= 1.0] = 1.0

        # Determine number of bins
        bin_number = np.max(discretised_voxels[mask_data])

        # Limit to maximum number of bins
        discretised_voxels[discretised_voxels >= bin_number * 1.0] = bin_number * 1.0

        # Set bin width.
        image.discretisation_bin_width = bin_width

        # Update IBSI compliance.
        image.ibsi_compliant = False

    else:
        raise ValueError(f"The discretisation_method argument was not recognised. Found: {discretisation_method}")

    # Update voxel grid.
    image.set_voxel_grid(discretised_voxels)
    image.discretisation_method = discretisation_method
    mask.intensity_range = tuple([1.0, bin_number])

    return image, mask



from typing import Generator

import pandas as pd

from mirp.settings.feature_parameters import FeatureExtractionSettingsClass
from mirp._features.base_feature import Feature
from mirp._features.morph_3d_features import generate_morph_3d_features
from mirp._features.local_intensity_features import generate_local_intensity_features
from mirp._features.ivh_features import generate_ivh_features
from mirp._features.stat_features import generate_stat_features
from mirp._features.ih_features import generate_ih_features
from mirp._features.cm_features import generate_cm_features
from mirp._features.dzm_features import generate_dzm_features
from mirp._features.rlm_features import generate_rlm_features
from mirp._features.szm_features import generate_szm_features
from mirp._features.ngtdm_features import generate_ngtdm_features
from mirp._features.ngldm_features import generate_ngldm_features


def generate_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[Feature, None, None]:

    # Morphological features
    yield from generate_morph_3d_features(settings=settings, features=features)

    # Local intensity features
    yield from generate_local_intensity_features(settings=settings, features=features)

    # Intensity-volume histogram
    yield from generate_ivh_features(settings=settings, features=features)

    # Statistical features
    yield from generate_stat_features(settings=settings, features=features)

    # Intensity histogram features
    yield from generate_ih_features(settings=settings, features=features)

    # Co-occurrence matrix features.
    yield from generate_cm_features(settings=settings, features=features)

    # Run length matrix features.
    yield from generate_rlm_features(settings=settings, features=features)

    # Size zone matrix features.
    yield from generate_szm_features(settings=settings, features=features)

    # Distance zone matrix features.
    yield from generate_dzm_features(settings=settings, features=features)

    # Neighbourhood grey tone difference matrix features.
    yield from generate_ngtdm_features(settings=settings, features=features)

    # Neighbouring grey level dependence matrix features.
    yield from generate_ngldm_features(settings=settings, features=features)


def feature_to_table(features: list[Feature]) -> pd.DataFrame | None:
    if features is None or len(features) == 0:
        return None

    # Set feature name for feature tables.
    for feature in features:
        feature.create_table_name()

    return pd.DataFrame(dict([(feature.table_name, [feature.value]) for feature in features]))



import copy
import numpy as np

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import MeanTransformedImage
from mirp._imagefilters.generic import GenericFilter
from mirp._imagefilters.utilities import SeparableFilterSet
from mirp.settings.generic import SettingsClass


class MeanFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = True
        self.ibsi_id = "S60F"

        # Set the filter size
        self.filter_size = settings.img_transform.mean_filter_size

        # Set the filter mode
        self.mode = settings.img_transform.mean_filter_boundary_condition

    def generate_object(self):
        # Generator for transformation objects.
        filter_size = copy.deepcopy(self.filter_size)
        if not isinstance(filter_size, list):
            filter_size = [filter_size]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_filter_size in filter_size:
            filter_object = copy.deepcopy(self)
            filter_object.filter_size = current_filter_size

            yield filter_object

    def transform(self, image: GenericImage) -> MeanTransformedImage:
        # Create placeholder Laws kernel response map.
        response_map = MeanTransformedImage(
            image_data=None,
            filter_size=self.filter_size,
            boundary_condition=self.mode,
            riesz_order=None,
            riesz_steering=None,
            riesz_sigma_parameter=None,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Set up the filter kernel.
        filter_kernel = np.ones(self.filter_size, dtype=float) / self.filter_size

        # Create a filter set.
        if self.separate_slices:
            filter_set = SeparableFilterSet(
                filter_x=filter_kernel,
                filter_y=filter_kernel
            )
        else:
            filter_set = SeparableFilterSet(
                filter_x=filter_kernel,
                filter_y=filter_kernel,
                filter_z=filter_kernel
            )

        # Apply the filter.
        response_map.set_voxel_grid(voxel_grid=filter_set.convolve(
            voxel_grid=image.get_voxel_grid(),
            mode=self.mode)
        )

        return response_map



from typing import Any
from dataclasses import dataclass
from mirp.settings.utilities import setting_def


@dataclass
class FeatureExtractionSettingsClass:
    """
    Parameters related to feature computation. Many are conditional on the type of features that will be computed (
    ``base_feature_families``). Features and their parameters are defined `here <https://arxiv.org/abs/1612.07003>`_.

    Parameters
    ----------
    by_slice: bool, optional, default: False
        Defines whether calculations should be performed in 2D (True) or 3D (False).
        See :class:`~mirp.settings.general_parameters.GeneralSettingsClass`.

    no_approximation: bool, optional, default: False
        Disables approximation of features, such as Geary's c-measure. Can be True or False (default). See
        :class:`~mirp.settings.general_parameters.GeneralSettingsClass`.

    ibsi_compliant: bool, optional, default: True
        Limits computation of features to those features that have a reference value in the IBSI reference standard.

    base_feature_families: str or list of str, optional, default: "none"
        Determines the feature families for which features are computed. Radiomics features are implemented as
        defined in the IBSI reference manual. The following feature families are currently present, and can be added
        using the following tags:

        * Morphological features: "mrp", "morph", "morphology", and "morphological".
        * Local intensity features: "li", "loc.int", "loc_int", "local_int", and "local_intensity".
        * Intensity-based statistical features: "st", "stat", "stats", "statistics", and "statistical".
        * Intensity histogram features: "ih", "int_hist", "int_histogram", and "intensity_histogram".
        * Intensity-volume histogram features: "ivh", "int_vol_hist", and "intensity_volume_histogram".
        * Grey level co-occurrence matrix (GLCM) features: "cm", "glcm", "grey_level_cooccurrence_matrix",
          and "cooccurrence_matrix".
        * Grey level run length matrix (GLRLM) features: "rlm", "glrlm", "grey_level_run_length_matrix", and
          "run_length_matrix".
        * Grey level size zone matrix (GLSZM) features: "szm", "glszm", "grey_level_size_zone_matrix", and
          "size_zone_matrix".
        * Grey level distance zone matrix (GLDZM) features: "dzm", "gldzm", "grey_level_distance_zone_matrix", and
          "distance_zone_matrix".
        * Neighbourhood grey tone difference matrix (NGTDM) features: "tdm", "ngtdm",
          "neighbourhood_grey_tone_difference_matrix", and "grey_tone_difference_matrix".
        * Neighbouring grey level dependence matrix (NGLDM) features: "ldm", "ngldm",
          "neighbouring_grey_level_dependence_matrix", and "grey_level_dependence_matrix".

        In addition, the following tags can be used:

        * "none": no features are computed.
        * "all": all features are computed.

        A list of strings may be provided to select multiple feature families.

        .. note::
            Even though ``"none"`` is the internal default, the :func:`~mirp.extractFeaturesAndImages.extract_features`
            function overrides this, and sets the default to ``"all"``.

    base_discretisation_method: {"fixed_bin_number", "fixed_bin_size", "fixed_bin_size_pyradiomics", "none"}
        Method used for discretising intensities. Used to compute intensity histogram as well as texture features.
        The setting is ignored if none of these feature families are being computed. The following options are
        available:

        * "fixed_bin_number": The intensity range within the mask is divided into a fixed number of bins,
          defined by the ``base_discretisation_bin_width`` parameter.
        * "fixed_bin_size": The intensity range is divided into bins with a fixed width, defined using the
          ``base_discretisation_bin_width`` parameter. The lower bound of the range is determined from the lower
          bound of the mask resegmentation range, see the ``resegmentation_intensity_range`` in
          :class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`. CT images have a default
          lower bound of the initial bin at -1000.0 and PET images have a default lower bound at 0.0. Other images,
          including MRI, normalised CT and PET images and filtered images, do not have a default value.
        * "fixed_bin_size_pyradiomics": The intensity range is divided into bins with a fixed width. This follows the
          non-IBSI compliant implementation in the pyradiomics package.
        * "none": The intensity range is not discretised into bins. This method can only be used if the image
          intensities are integer and strictly positive.

         There is no default method. Multiple methods can be specified as a list to yield features according to each
         method.

        .. warning::
            The "fixed_bin_size_pyradiomics" is not IBSI compliant, and should only be used when
            reproducing results from studies that used pyradiomics.

    base_discretisation_n_bins: int or list of int
        Number of bins used for the "fixed_bin_number" discretisation method. No default value. Multiple values can
        be specified in a list to yield features according to each number of bins.

    base_discretisation_bin_width: float or list of float
        Width of each bin in the "fixed_bin_size" discretisation method. No default value. Multiple values can be
        specified in a list to yield features according to each bin width.

    ivh_discretisation_method: {"fixed_bin_number", "fixed_bin_size", "none"}, optional, default: "none"
        Method used for discretising intensities for computing intensity-volume histograms. The discretisation
        methods follow those in ``base_discretisation_method``. The "none" method changes to "fixed_bin_number" if
        the underlying data are not suitable.

    ivh_discretisation_n_bins: int, optional, default: 1000
        Number of bins used for the "fixed_bin_number" discretisation method.

    ivh_discretisation_bin_width: float, optional
        Width of each bin in the "fixed_bin_size" discretisation method. No default value.

    glcm_distance: float or list of float, optional, default: 1.0
        Distance (in voxels) for GLCM for determining the neighbourhood. Chebyshev, or checkerboard, distance is
        used. A value of 1.0 will therefore consider all (diagonally) adjacent voxels as its neighbourhood. A list of
        values can be provided to compute GLCM features at different scales.

    glcm_spatial_method: {"2d_average", "2d_slice_merge", "2.5d_direction_merge", "2.5d_volume_merge", "3d_average", "3d_volume_merge"}, optional
        Determines how co-occurrence matrices are formed and aggregated. One of the following:

        * "2d_average": features are computed from all matrices then averaged [IBSI:BTW3].
        * "2d_slice_merge": matrices in the same slice are merged, features computed and then averaged [IBSI:SUJT].
        * "2.5d_direction_merge": matrices for the same direction are merged, features computed and then averaged
          [IBSI:JJUI].
        * "2.5d_volume_merge": all matrices are merged and a single feature is calculated [IBSI:ZW7Z].
        * "3d_average": features are computed from all matrices then averaged [IBSI:ITBB].
        * "3d_volume_merge": all matrices are merged and a single feature is computed from the merged matrix
          [IBSI:IAZD].

        A list of values may be provided to extract features for multiple spatial methods. Default: "2d_slice_merge"
        (``by_slice = True``) or "3d_volume_merge" (``by_slice = False``).

    glrlm_spatial_method: {"2d_average", "2d_slice_merge", "2.5d_direction_merge", "2.5d_volume_merge", "3d_average", "3d_volume_merge"}, optional
        Determines how run length matrices are formed and aggregated. One of the following:

        * "2d_average": features are calculated from all matrices then averaged [IBSI:BTW3].
        * "2d_slice_merge": matrices in the same slice are merged, features computed and then averaged [IBSI:SUJT].
        * "2.5d_direction_merge": matrices for the same direction are merged, features computed and then averaged
          [IBSI:JJUI].
        * "2.5d_volume_merge": all matrices are merged and a single feature is computed [IBSI:ZW7Z].
        * "3d_average": features are computed from all matrices then averaged [IBSI:ITBB].
        * "3d_volume_merge": all matrices are merged and a single feature is computed from the merged matrix
          [IBSI:IAZD].

        A list of values may be provided to extract features for multiple spatial methods. Default:
        "2d_slice_merge" (``by_slice = True``) or "3d_volume_merge" (``by_slice = False``).

    glszm_spatial_method: {"2d", "2.5d", "3d"}, optional
        Determines how the size zone matrices are formed and aggregated. One of the following:

        * "2d": features are computed from individual matrices and subsequently averaged [IBSI:8QNN].
        * "2.5d": all 2D matrices are merged and features are computed from this single matrix [IBSI:62GR].
        * "3d": features are computed from a single 3D matrix [IBSI:KOBO].

        A list of values may be provided to extract features for multiple spatial methods. Default: "2d"
        (``by_slice = True``) or "3d" (``by_slice = False``).

    gldzm_spatial_method: {"2d", "2.5d", "3d"}, optional
        Determines how the distance zone matrices are formed and aggregated. One of the following:

        * "2d": features are computed from individual matrices and subsequently averaged [IBSI:8QNN].
        * "2.5d": all 2D matrices are merged and features are computed from this single matrix [IBSI:62GR].
        * "3d": features are computed from a single 3D matrix [IBSI:KOBO].

        A list of values may be provided to extract features for multiple spatial methods. Default: "2d"
        (``by_slice = True``) or "3d" (``by_slice = False``).

    ngtdm_spatial_method: {"2d", "2.5d", "3d"}, optional
        Determines how the neighbourhood grey tone difference matrices are formed and aggregated. One of the
        following:

        * "2d": features are computed from individual matrices and subsequently averaged [IBSI:8QNN].
        * "2.5d": all 2D matrices are merged and features are computed from this single matrix [IBSI:62GR].
        * "3d": features are computed from a single 3D matrix [IBSI:KOBO].

        A list of values may be provided to extract features for multiple spatial methods. Default: "2d"
        (``by_slice = True``) or "3d" (``by_slice = False``).

    ngldm_distance: float or list of float, optional, default: 1.0
        Distance (in voxels) for NGLDM for determining the neighbourhood. Chebyshev, or checkerboard, distance is
        used. A value of 1.0 will therefore consider all (diagonally) adjacent voxels as its neighbourhood. A list of
        values can be provided to compute NGLDM features at different scales.

    ngldm_difference_level: float or list of float, optional, default: 0.0
        Difference level (alpha) for NGLDM. Determines which bins are grouped together in the matrix.

    ngldm_spatial_method: {"2d", "2.5d", "3d"}, optional
        Determines how the neighbourhood grey level dependence matrices are formed and aggregated. One of the
        following:

        * "2d": features are computed from individual matrices and subsequently averaged [IBSI:8QNN].
        * "2.5d": all 2D matrices are merged and features are computed from this single matrix [IBSI:62GR].
        * "3d": features are computed from a single 3D matrix [IBSI:KOBO].

        A list of values may be provided to extract features for multiple spatial methods. Default: "2d"
        (``by_slice = True``) or "3d" (``by_slice = False``).

    **kwargs: dict, optional
        Unused keyword arguments.
    """

    def __init__(
            self,
            by_slice: bool = False,
            no_approximation: bool = False,
            ibsi_compliant: bool = True,
            base_feature_families: None | str | list[str] = "none",
            base_discretisation_method: None | str | list[str] = None,
            base_discretisation_n_bins: None | int | list[int] = None,
            base_discretisation_bin_width: None | float | list[float] = None,
            ivh_discretisation_method: str = "none",
            ivh_discretisation_n_bins: None | int = 1000,
            ivh_discretisation_bin_width: None | float = None,
            glcm_distance: float | list[float] = 1.0,
            glcm_spatial_method: None | str | list[str] = None,
            glrlm_spatial_method: None | str | list[str] = None,
            glszm_spatial_method: None | str | list[str] = None,
            gldzm_spatial_method: None | str | list[str] = None,
            ngtdm_spatial_method: None | str | list[str] = None,
            ngldm_distance: float | list[float] = 1.0,
            ngldm_difference_level: float | list[float] = 0.0,
            ngldm_spatial_method: None | str | list[str] = None,
            **kwargs
    ):
        # Set by slice.
        self.by_slice: bool = by_slice

        # Set approximation flag.
        self.no_approximation: bool = no_approximation

        # Set IBSI-compliance flag.
        self.ibsi_compliant: bool = ibsi_compliant

        if base_feature_families is None:
            base_feature_families = "none"

        # Check families.
        if not isinstance(base_feature_families, list):
            base_feature_families = [base_feature_families]

        # Check which entries are valid.
        valid_families: list[bool] = [ii in self.get_available_families() for ii in base_feature_families]

        if not all(valid_families):
            raise ValueError(
                f"One or more families in the base_feature_families parameter were not recognised: "
                f"{', '.join([base_feature_families[ii] for ii, is_valid in enumerate(valid_families) if not is_valid])}")

        # Set families.
        self.families: list[str] = base_feature_families

        if not self.has_any_feature_family():
            self.families = ["none"]

        if self.has_discretised_family():
            # Check if discretisation_method is None.
            if base_discretisation_method is None:
                raise ValueError("The base_discretisation_method parameter has no default and must be set.")

            if not isinstance(base_discretisation_method, list):
                base_discretisation_method = [base_discretisation_method]

            if not all(discretisation_method in [
                "fixed_bin_size", "fixed_bin_number", "fixed_bin_size_pyradiomics", "none"
            ] for discretisation_method in base_discretisation_method):
                raise ValueError(
                    "Available values for the base_discretisation_method parameter are "
                    "'fixed_bin_number', 'fixed_bin_size', 'fixed_bin_size_pyradiomics' and 'none'. "
                    "One or more values were not recognised."
                )

            # Check discretisation_n_bins
            if "fixed_bin_number" in base_discretisation_method:
                if base_discretisation_n_bins is None:
                    raise ValueError("The base_discretisation_n_bins parameter has no default and must be set")

                if not isinstance(base_discretisation_n_bins, list):
                    base_discretisation_n_bins = [base_discretisation_n_bins]

                if not all(isinstance(n_bins, int) for n_bins in base_discretisation_n_bins):
                    raise TypeError(
                        "The base_discretisation_n_bins parameter is expected to contain integers with "
                        "value 2 or larger. Found one or more values that were not integers.")

                if not all(n_bins >= 2 for n_bins in base_discretisation_n_bins):
                    raise ValueError(
                        "The base_discretisation_n_bins parameter is expected to contain integers with "
                        "value 2 or larger. Found one or more values that were less than 2.")

            else:
                base_discretisation_n_bins = None

            # Check discretisation_bin_width
            if "fixed_bin_size" in base_discretisation_method or "fixed_bin_size_pyradiomics" in base_discretisation_method:
                if base_discretisation_bin_width is None:
                    raise ValueError(
                        "The base_discretisation_bin_width parameter has no default value and must be set.")

                if not isinstance(base_discretisation_bin_width, list):
                    base_discretisation_bin_width = [base_discretisation_bin_width]

                if not all(isinstance(bin_size, float) for bin_size in base_discretisation_bin_width):
                    raise TypeError(
                        "The base_discretisation_bin_width parameter is expected to contain floating "
                        "point values greater than 0.0. Found one or more values that were not floating "
                        "points.")

                if not all(bin_size > 0.0 for bin_size in base_discretisation_bin_width):
                    raise ValueError(
                        "The base_discretisation_bin_width parameter is expected to contain floating "
                        "point values greater than 0.0. Found one or more values that were 0.0 or less.")

                # Check ibsi_compliance
                if self.ibsi_compliant and "fixed_bin_size_pyradiomics" in base_discretisation_method:
                    raise ValueError(
                        "The fixed_bin_size_pyradiomics method is not IBSI-compliant. If you are sure that you want to "
                        "use this method, use ibsi_compliant = False."
                    )

            else:
                base_discretisation_bin_width = None

        else:
            base_discretisation_method = None
            base_discretisation_n_bins = None
            base_discretisation_bin_width = None

        # Set discretisation method-related parameters.
        self.discretisation_method: None | list[str] = base_discretisation_method
        self.discretisation_n_bins: None | list[int] = base_discretisation_n_bins
        self.discretisation_bin_width: None | list[float] = base_discretisation_bin_width

        if self.has_ivh_family():
            if ivh_discretisation_method not in ["fixed_bin_size", "fixed_bin_number", "none"]:
                raise ValueError(
                    "Available values for the ivh_discretisation_method parameter are 'fixed_bin_size', "
                    "'fixed_bin_number', and 'none'. One or more values were not recognised.")

            # Check discretisation_n_bins
            if "fixed_bin_number" in ivh_discretisation_method:

                if not isinstance(ivh_discretisation_n_bins, int):
                    raise TypeError(
                        "The ivh_discretisation_n_bins parameter is expected to be an integer with "
                        "value 2 or greater. Found: a value that was not an integer.")

                if not ivh_discretisation_n_bins >= 2:
                    raise ValueError(
                        "The ivh_discretisation_n_bins parameter is expected to be an integer with "
                        f"value 2 or greater. Found: {ivh_discretisation_n_bins}")

            else:
                ivh_discretisation_n_bins = None

            # Check discretisation_bin_width
            if "fixed_bin_size" in ivh_discretisation_method:

                if not isinstance(ivh_discretisation_bin_width, float):
                    raise TypeError(
                        "The ivh_discretisation_bin_width parameter is expected to be a floating "
                        "point value greater than 0.0. Found a value that was not a floating point.")

                if not ivh_discretisation_bin_width > 0.0:
                    raise ValueError(
                        "The ivh_discretisation_bin_width parameter is expected to  be a floating "
                        f"point value greater than 0.0. Found: {ivh_discretisation_bin_width}")

            else:
                ivh_discretisation_bin_width = None

        else:
            ivh_discretisation_method = None
            ivh_discretisation_n_bins = None
            ivh_discretisation_bin_width = None

        # Set parameters
        self.ivh_discretisation_method: None | str = ivh_discretisation_method
        self.ivh_discretisation_n_bins: None | int = ivh_discretisation_n_bins
        self.ivh_discretisation_bin_width: None | float = ivh_discretisation_bin_width

        # Set GLCM attributes.
        if self.has_glcm_family():
            # Check distance parameter.
            if not isinstance(glcm_distance, list):
                glcm_distance = [glcm_distance]

            if not all(isinstance(distance, float) for distance in glcm_distance):
                raise TypeError(
                    "The glcm_distance parameter is expected to contain floating point values of 1.0 "
                    "or greater. Found one or more values that were not floating points.")

            if not all(distance >= 1.0 for distance in glcm_distance):
                raise ValueError(
                    "The glcm_distance parameter is expected to contain floating point values of 1.0 "
                    "or greater. Found one or more values that were less than 1.0.")

            # Check spatial method.
            glcm_spatial_method = self.check_valid_directional_spatial_method(
                glcm_spatial_method,
                "glcm_spatial_method")

        else:
            glcm_distance = None
            glcm_spatial_method = None

        self.glcm_distance: None | list[float] = glcm_distance
        self.glcm_spatial_method: None | list[str] = glcm_spatial_method

        # Set GLRLM attributes.
        if self.has_glrlm_family():
            # Check spatial method.
            glrlm_spatial_method = self.check_valid_directional_spatial_method(
                glrlm_spatial_method, "glrlm_spatial_method")

        else:
            glrlm_spatial_method = None

        self.glrlm_spatial_method: None | list[str] = glrlm_spatial_method

        # Set GLSZM attributes.
        if self.has_glszm_family():
            # Check spatial method.
            glszm_spatial_method = self.check_valid_omnidirectional_spatial_method(
                glszm_spatial_method, "glszm_spatial_method")
        else:
            glszm_spatial_method = None

        self.glszm_spatial_method: None | list[str] = glszm_spatial_method

        # Set GLDZM attributes.
        if self.has_gldzm_family():
            # Check spatial method.
            gldzm_spatial_method = self.check_valid_omnidirectional_spatial_method(
                gldzm_spatial_method, "gldzm_spatial_method")

        else:
            gldzm_spatial_method = None

        self.gldzm_spatial_method: None | list[str] = gldzm_spatial_method

        # Set NGTDM attributes.
        if self.has_ngtdm_family():
            # Check spatial method
            ngtdm_spatial_method = self.check_valid_omnidirectional_spatial_method(
                ngtdm_spatial_method, "ngtdm_spatial_method")

        else:
            ngtdm_spatial_method = None

        self.ngtdm_spatial_method: None | list[str] = ngtdm_spatial_method

        # Set NGLDM attributes
        if self.has_ngldm_family():

            # Check distance.
            if not isinstance(ngldm_distance, list):
                ngldm_distance = [ngldm_distance]

            if not all(isinstance(distance, float) for distance in ngldm_distance):
                raise TypeError(
                    "The ngldm_distance parameter is expected to contain floating point values of 1.0 "
                    "or greater. Found one or more values that were not floating points.")

            if not all(distance >= 1.0 for distance in ngldm_distance):
                raise ValueError(
                    "The ngldm_distance parameter is expected to contain floating point values of 1.0 "
                    "or greater. Found one or more values that were less than 1.0.")

            # Check spatial method
            ngldm_spatial_method = self.check_valid_omnidirectional_spatial_method(
                ngldm_spatial_method, "ngldm_spatial_method")

            # Check difference level.
            if not isinstance(ngldm_difference_level, list):
                ngldm_difference_level = [ngldm_difference_level]

            if not all(isinstance(difference, float) for difference in ngldm_difference_level):
                raise TypeError(
                    "The ngldm_difference_level parameter is expected to contain floating point values of 0.0 "
                    "or greater. Found one or more values that were not floating points.")

            if not all(difference >= 0.0 for difference in ngldm_difference_level):
                raise ValueError(
                    "The ngldm_difference_level parameter is expected to contain floating point values "
                    "of 0.0 or greater. Found one or more values that were less than 0.0.")

        else:
            ngldm_spatial_method = None
            ngldm_distance = None
            ngldm_difference_level = None

        self.ngldm_dist: None | list[float] = ngldm_distance
        self.ngldm_diff_lvl: None | list[float] = ngldm_difference_level
        self.ngldm_spatial_method: None | list[str] = ngldm_spatial_method

    @staticmethod
    def get_available_families():
        return [
            "mrp", "morph", "morphology", "morphological", "li", "loc.int", "loc_int", "local_int", "local_intensity",
            "st", "stat", "stats", "statistics", "statistical", "ih", "int_hist", "int_histogram", "intensity_histogram",
            "ivh", "int_vol_hist", "intensity_volume_histogram", "cm", "glcm", "grey_level_cooccurrence_matrix",
            "cooccurrence_matrix", "rlm", "glrlm", "grey_level_run_length_matrix", "run_length_matrix",
            "szm", "glszm", "grey_level_size_zone_matrix", "size_zone_matrix", "dzm", "gldzm",
            "grey_level_distance_zone_matrix", "distance_zone_matrix", "tdm", "ngtdm",
            "neighbourhood_grey_tone_difference_matrix", "grey_tone_difference_matrix", "ldm", "ngldm",
            "neighbouring_grey_level_dependence_matrix", "grey_level_dependence_matrix", "all", "none"
        ]

    def has_any_feature_family(self):
        return not any(family == "none" for family in self.families)

    def has_discretised_family(self):
        return self.has_ih_family() or self.has_glcm_family() or self.has_glrlm_family() or self.has_glszm_family() \
               or self.has_gldzm_family() or self.has_ngtdm_family() or self.has_ngldm_family()

    def has_morphology_family(self):
        return any(family in ["mrp", "morph", "morphology", "morphological", "all"] for family in self.families)

    def has_local_intensity_family(self):
        return any(family in ["li", "loc.int", "loc_int", "local_int", "local_intensity", "all"] for family in self.families)

    def has_stats_family(self):
        return any(family in ["st", "stat", "stats", "statistics", "statistical", "all"] for family in self.families)

    def has_ih_family(self):
        return any(family in ["ih", "int_hist", "int_histogram", "intensity_histogram", "all"] for family in self.families)

    def has_ivh_family(self):
        return any(family in ["ivh", "int_vol_hist", "intensity_volume_histogram", "all"] for family in self.families)

    def has_glcm_family(self):
        return any(family in ["cm", "glcm", "grey_level_cooccurrence_matrix", "cooccurrence_matrix", "all"] for family in self.families)

    def has_glrlm_family(self):
        return any(family in ["rlm", "glrlm", "grey_level_run_length_matrix", "run_length_matrix", "all"] for family in self.families)

    def has_glszm_family(self):
        return any(family in ["szm", "glszm", "grey_level_size_zone_matrix", "size_zone_matrix", "all"] for family in self.families)

    def has_gldzm_family(self):
        return any(family in ["dzm", "gldzm", "grey_level_distance_zone_matrix", "distance_zone_matrix", "all"] for family in self.families)

    def has_ngtdm_family(self):
        return any(family in ["tdm", "ngtdm", "neighbourhood_grey_tone_difference_matrix", "grey_tone_difference_matrix", "all"] for family in self.families)

    def has_ngldm_family(self):
        return any(family in ["ldm", "ngldm", "neighbouring_grey_level_dependence_matrix", "grey_level_dependence_matrix", "all"] for family in self.families)

    def check_valid_directional_spatial_method(self, x, var_name):

        # Set defaults
        if x is None and self.by_slice:
            x = ["2d_slice_merge"]

        elif x is None and not self.by_slice:
            x = ["3d_volume_merge"]

        # Check that x is a list.
        if not isinstance(x, list):
            x = [x]

        all_spatial_method = ["2d_average", "2d_slice_merge", "2.5d_direction_merge", "2.5d_volume_merge"]
        if not self.by_slice:
            all_spatial_method += ["3d_average", "3d_volume_merge"]

        # Check that x contains strings.
        if not all(isinstance(spatial_method, str) for spatial_method in x):
            raise TypeError(
                f"The {var_name} parameter expects one or more of the following values: "
                f"{', '.join(all_spatial_method)}. Found: one or more values that were not strings.")

        # Check spatial method.
        valid_spatial_method = [spatial_method in all_spatial_method for spatial_method in x]

        if not all(valid_spatial_method):
            raise ValueError(
                f"The {var_name} parameter expects one or more of the following values: "
                f"{', '.join(all_spatial_method)}. Found: "
                f"{', '.join([spatial_method for spatial_method in x if spatial_method in all_spatial_method])}")

        return x

    def check_valid_omnidirectional_spatial_method(self, x, var_name):

        # Set defaults
        if x is None and self.by_slice:
            x = ["2d"]

        elif x is None and not self.by_slice:
            x = ["3d"]

        # Check that x is a list.
        if not isinstance(x, list):
            x = [x]

        all_spatial_method = ["2d", "2.5d"]
        if not self.by_slice:
            all_spatial_method += ["3d"]

        # Check that x contains strings.
        if not all(isinstance(spatial_method, str) for spatial_method in x):
            raise TypeError(
                f"The {var_name} parameter expects one or more of the following values: "
                f"{', '.join(all_spatial_method)}. Found: one or more values that were not strings.")

        # Check spatial method.
        valid_spatial_method = [spatial_method in all_spatial_method for spatial_method in x]

        if not all(valid_spatial_method):
            raise ValueError(
                f"The {var_name} parameter expects one or more of the following values: "
                f"{', '.join(all_spatial_method)}. Found: "
                f"{', '.join([spatial_method for spatial_method in x if spatial_method in all_spatial_method])}")

        return x


def get_feature_extraction_settings() -> list[dict[str, Any]]:
    return [
        setting_def(
            "base_feature_families", "str", to_list=True, xml_key=["feature_families", "families"],
            class_key="families", test=["all"]
        ),
        setting_def(
            "base_discretisation_method", "str", to_list=True, xml_key=["discretisation_method", "discr_method"],
            class_key="discretisation_method", test=["fixed_bin_size", "fixed_bin_number"]
        ),
        setting_def(
            "base_discretisation_n_bins", "int", to_list=True, xml_key=["discretisation_n_bins", "discr_n_bins"],
            class_key="discretisation_n_bins", test=[10, 33]
        ),
        setting_def(
            "base_discretisation_bin_width", "float", to_list=True,
            xml_key=["discretisation_bin_width", "discr_bin_width"], class_key="discretisation_bin_width",
            test=[10.0, 34.0]
        ),
        setting_def(
            "ivh_discretisation_method", "str", xml_key=["ivh_discretisation_method", "ivh_discr_method"],
            class_key="ivh_discretisation_method", test="fixed_bin_size"
        ),
        setting_def(
            "ivh_discretisation_n_bins", "int", xml_key=["ivh_discretisation_n_bins", "ivh_discr_n_bins"],
            test=20
        ),
        setting_def(
            "ivh_discretisation_bin_width", "float", xml_key=["ivh_discretisation_bin_width", "ivh_discr_bin_width"],
            test=30.0
        ),
        setting_def("glcm_distance", "float", to_list=True, xml_key=["glcm_distance", "glcm_dist"], test=[2.0, 3.0]),
        setting_def("glcm_spatial_method", "str", to_list=True, test=["2d_average", "2d_slice_merge"]),
        setting_def("glrlm_spatial_method", "str", to_list=True, test=["2d_average", "2d_slice_merge"]),
        setting_def("glszm_spatial_method", "str", to_list=True, test=["2d", "2.5d"]),
        setting_def("gldzm_spatial_method", "str", to_list=True, test=["2d", "2.5d"]),
        setting_def("ngtdm_spatial_method", "str", to_list=True, test=["2d", "2.5d"]),
        setting_def(
            "ngldm_distance", "float", to_list=True, xml_key=["ngldm_distance", "ngldm_dist"],
            class_key="ngldm_dist", test=[2.5, 3.5]
        ),
        setting_def(
            "ngldm_difference_level", "float", to_list=True, xml_key=["ngldm_difference_level", "ngldm_diff_lvl"],
            class_key="ngldm_diff_lvl", test=[1.0, 1.9]
        ),
        setting_def("ngldm_spatial_method", "str", to_list=True, test=["2d", "2.5d"])
    ]



import copy
import numpy as np

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import SquareTransformedImage
from mirp._imagefilters.generic import GenericFilter
from mirp.settings.generic import SettingsClass


class SquareTransformFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        # Square transform filters are not IBSI-compliant.
        self.ibsi_compliant: bool = False

    def generate_object(self):
        yield copy.deepcopy(self)

    def transform(self, image: GenericImage) -> SquareTransformedImage:
        # Create placeholder response map.
        response_map = SquareTransformedImage(
            image_data=None,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        image_data = image.get_voxel_grid()
        alpha = np.sqrt(np.max(np.abs(image_data)))

        # Prevent issues with alpha values that are not strictly positive.
        if not np.isfinite(alpha) or alpha == 0.0:
            alpha = 1.0

        response_map.set_voxel_grid(
            voxel_grid=np.power(image_data / alpha, 2.0)
        )

        return response_map



from mirp.settings.generic import SettingsClass
from mirp._images.generic_image import GenericImage


class GenericFilter:

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):
        # In-slice (2D) or 3D filtering
        self.separate_slices = image.separate_slices

        # Even though most currently implemented filters are IBSI-compliant, set value to False to avoid surprises in
        # the future.
        self.ibsi_compliant: bool = False
        self.ibsi_id: None | str = None

    def generate_object(self):
        raise NotImplementedError("_generate_object method should be defined in the subclasses")

    def transform(self, image: GenericImage):
        raise NotImplementedError("transform method should be defined in the subclasses.")



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._features.rlm_matrix import MatrixRLM
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters
from mirp._features.texture_features import FeatureTexture
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class FeatureRLM(FeatureTexture):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Perform close crop for RLM.
        self.cropping_distance = 0.0

    def get_matrix(
            self,
            image: GenericImage,
            mask: BaseMask,
    ) -> list[MatrixRLM]:
        # First discretise the image.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Then get matrix or matrices
        matrix_list = self._get_matrix(
            image=image,
            mask=mask,
            spatial_method=self.spatial_method
        )

        return matrix_list

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_matrix(
            image: GenericImage,
            mask: BaseMask,
            spatial_method: str
    ) -> list[MatrixRLM]:
        # Represent image and mask as a dataframe.
        data = mask.as_pandas_dataframe(
            image=image,
            intensity_mask=True
        )

        # Instantiate a helper copy of the current class to be able to use class methods without tying the cache to the
        # instance of the original object from which this method is called.
        matrix_instance = MatrixRLM(
            spatial_method=spatial_method
        )

        # Compute the required matrices.
        matrix_list = list(matrix_instance.generate(prototype=MatrixRLM, n_slices=image.image_dimension[0]))
        for matrix in matrix_list:
            matrix.compute(data=data, image_dimension=image.image_dimension)

        # Merge according to the spatial method.
        matrix_list = matrix_instance.merge(matrix_list, prototype=MatrixRLM)

        # Compute additional values from the individual matrices.
        for matrix in matrix_list:
            matrix.set_values_from_matrix()

        return matrix_list

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureRLM):
            self._get_matrix.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_matrix.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Compute or retrieve matrices from cache.
        matrices = self.get_matrix(image=image, mask=mask)

        # Compute feature value from matrices, and average over matrices.
        values = [self._compute(matrix=matrix) for matrix in matrices]
        if np.all(np.isnan(values)):
            self.value = np.nan
        else:
            self.value = np.nanmean(values)

    @staticmethod
    def _compute(matrix: MatrixRLM):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + self._get_spatial_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureRLMSRE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - short runs emphasis"
        self.abbr_name = "rlm_sre"
        self.ibsi_id = "22OV"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rj.rj / matrix.rj.j ** 2.0) / matrix.n_s


class FeatureRLMLRE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - long runs emphasis"
        self.abbr_name = "rlm_lre"
        self.ibsi_id = "W4KF"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rj.rj * matrix.rj.j ** 2.0) / matrix.n_s


class FeatureRLMLGRE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - low grey level run emphasis"
        self.abbr_name = "rlm_lgre"
        self.ibsi_id = "V3SW"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.ri.ri / matrix.ri.i ** 2.0) / matrix.n_s


class FeatureRLMHGRE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - high grey level run emphasis"
        self.abbr_name = "rlm_hgre"
        self.ibsi_id = "G3QZ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.ri.ri * matrix.ri.i ** 2.0) / matrix.n_s


class FeatureRLMSRLGE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - short run low grey level emphasis"
        self.abbr_name = "rlm_srlge"
        self.ibsi_id = "HTZT"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rij.rij / (matrix.rij.i * matrix.rij.j) ** 2.0) / matrix.n_s


class FeatureRLMSRHGE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - short run high grey level emphasis"
        self.abbr_name = "rlm_srhge"
        self.ibsi_id = "GD3A"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rij.rij * matrix.rij.i ** 2.0 / matrix.rij.j ** 2.0) / matrix.n_s


class FeatureRLMLRLGE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - long run low grey level emphasis"
        self.abbr_name = "rlm_lrlge"
        self.ibsi_id = "IVPO"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rij.rij * matrix.rij.j ** 2.0 / matrix.rij.i ** 2.0) / matrix.n_s


class FeatureRLMLRHGE(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - long run high grey level emphasis"
        self.abbr_name = "rlm_lrhge"
        self.ibsi_id = "3KUM"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rij.rij * matrix.rij.i ** 2.0 * matrix.rij.j ** 2.0) / matrix.n_s


class FeatureRLMGLNU(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - grey level non-uniformity"
        self.abbr_name = "rlm_glnu"
        self.ibsi_id = "R5YN"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.ri.ri ** 2.0) / matrix.n_s


class FeatureRLMGLNUNorm(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - normalised grey level non-uniformity"
        self.abbr_name = "rlm_glnu_norm"
        self.ibsi_id = "OVBL"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.ri.ri ** 2.0) / matrix.n_s ** 2.0


class FeatureRLMRLNU(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - run length non-uniformity)"
        self.abbr_name = "rlm_rlnu"
        self.ibsi_id = "W92Y"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rj.rj ** 2.0) / matrix.n_s


class FeatureRLMRLNUNorm(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - normalised run length non-uniformity"
        self.abbr_name = "rlm_rlnu_norm"
        self.ibsi_id = "IC23"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.rj.rj ** 2.0) / matrix.n_s ** 2.0


class FeatureRLMRPerc(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - run percentage"
        self.abbr_name = "rlm_r_perc"
        self.ibsi_id = "9ZK5"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return 1.0 * matrix.n_s / matrix.n_voxels


class FeatureRLMGLVar(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - grey level variance"
        self.abbr_name = "rlm_gl_var"
        self.ibsi_id = "8CE5"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.rij.rij * matrix.rij.i) / matrix.n_s
        return np.sum((matrix.rij.i - mu) ** 2.0 * matrix.rij.rij) / matrix.n_s


class FeatureRLMRLVar(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - run length variance"
        self.abbr_name = "rlm_rl_var"
        self.ibsi_id = "SXLW"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.rij.rij * matrix.rij.j) / matrix.n_s
        return np.sum((matrix.rij.j - mu) ** 2.0 * matrix.rij.rij) / matrix.n_s


class FeatureRLMRLEntr(FeatureRLM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "RLM - run entropy"
        self.abbr_name = "rlm_rl_entr"
        self.ibsi_id = "HJ9O"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixRLM) -> float:
        if matrix.is_empty():
            return np.nan
        return - np.sum(matrix.rij.rij * np.log2(matrix.rij.rij / matrix.n_s)) / matrix.n_s


def get_rlm_class_dict() -> dict[str, FeatureRLM]:
    class_dict = {
        "rlm_sre": FeatureRLMSRE,
        "rlm_lre": FeatureRLMLRE,
        "rlm_lgre": FeatureRLMLGRE,
        "rlm_hgre": FeatureRLMHGRE,
        "rlm_srlge": FeatureRLMSRLGE,
        "rlm_srhge": FeatureRLMSRHGE,
        "rlm_lrlge": FeatureRLMLRLGE,
        "rlm_lrhge": FeatureRLMLRHGE,
        "rlm_glnu": FeatureRLMGLNU,
        "rlm_glnu_norm": FeatureRLMGLNUNorm,
        "rlm_rlnu": FeatureRLMRLNU,
        "rlm_rlnu_norm": FeatureRLMRLNUNorm,
        "rlm_r_perc": FeatureRLMRPerc,
        "rlm_gl_var": FeatureRLMGLVar,
        "rlm_rl_var": FeatureRLMRLVar,
        "rlm_rl_entr": FeatureRLMRLEntr
    }

    return class_dict


def generate_rlm_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureRLM, None, None]:
    class_dict = get_rlm_class_dict()
    rlm_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_glrlm_family():
        features = rlm_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only RLM-features, and return if none are present.
    features = [feature for feature in features if feature in rlm_features]
    if len(features) == 0:
        return

    # Features are parametrised by the choice of discretisation parameters and spatial methods.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for spatial_method in settings.glrlm_spatial_method:
            for feature in features:
                yield class_dict[feature](
                    spatial_method=spatial_method,
                    **discretisation_parameters
                )



from typing import Iterable, Any
from dataclasses import dataclass
from mirp.settings.utilities import setting_def


@dataclass
class ImageInterpolationSettingsClass:
    """
    Parameters related to image interpolating / resampling. Images in a dataset are typically resampled to uniform
    voxel spacing to ensure that their spatial representation does not vary between samples.

    For parameters related to mask interpolation / resampling, see
    :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`.

    Parameters
    ----------
    by_slice: bool, optional, default: False
        Defines whether calculations should be performed in 2D (True) or 3D (False).
        See :class:`~mirp.settings.general_parameters.GeneralSettingsClass`.

    new_spacing: float or list of float or list of list of float, optional:
        Sets voxel spacing after interpolation. A single value represents the spacing that will be applied in all
        directions. Non-uniform voxel spacing may also be provided, but requires 3 values for z, y, and x directions
        (if `by_slice = False`) or 2 values for y and x directions (otherwise).

        Multiple spacings may be defined by creating a nested list, e.g. [[1.0], [1.5], [2.0]] to resample the
        same image multiple times to different (here: isotropic) voxel spacings, namely 1.0, 1.5 and 2.0. Units
        are defined by the headers of the image files. These are typically millimeters for radiological images.

    spline_order: int, optional, default: 3
        Sets the spline order used for spline interpolation. mirp uses `scipy.ndimage.map_coordinates
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.map_coordinates.html#scipy.ndimage
        .map_coordinates>`_ internally. Spline orders 0, 1, and 3 refer to nearest neighbour, linear interpolation
        and cubic interpolation, respectively.

    anti_aliasing: bool, optional, default: true
        Determines whether to perform antialiasing, which is done to mitigate aliasing artifacts when downsampling.

        .. note::
            When voxel spacing in the original image is smaller than that in the resampled image (e.g., 0.5 mm sampled
            to 1.0 mm), antialiasing is recommended `[Mackin et al.] <http://dx.doi.org/10.1371/journal.pone.0178524>`_.

    smoothing_beta: float, optional, default: 0.98
        Determines the smoothness of the Gaussian filter used for anti-aliasing. A value of 1.00 equates to no
        antialiasing, with lower values producing increasingly smooth imaging. Values above 0.90 are recommended.
        The effect of this parameter is shown in the supplement of `Zwanenburg et al.
        <http://dx.doi.org/10.1038/s41598-018-36938-4>`_.

    **kwargs: dict, optional
        Unused keyword arguments.
    """

    def __init__(
            self,
            by_slice: bool,
            new_spacing: None | float | int | list[int] | list[float] | list[list[float]] | list[list[int]] = None,
            spline_order: int = 3,
            anti_aliasing: bool = True,
            smoothing_beta: float = 0.98,
            **kwargs
    ):

        # Set interpolate parameter.
        self.interpolate: bool = new_spacing is not None

        # Check if the spline order is valid.
        if spline_order < 0 or spline_order > 5:
            raise ValueError(
                f"The interpolation spline order should be an integer between 0 and 5. Found: {spline_order}")

        # Set spline order for the interpolating spline.
        self.spline_order: int = spline_order

        if self.interpolate:
            # Parse value to list of floating point values to facilitate checks.
            if isinstance(new_spacing, (int, float)):
                new_spacing = [new_spacing]

            # Check if nested list elements are present.
            if any(isinstance(ii, Iterable) for ii in new_spacing):
                new_spacing = [
                    self._check_new_sample_spacing(by_slice=by_slice, new_spacing=new_spacing_element)
                    for new_spacing_element in new_spacing
                ]

            else:
                new_spacing = [self._check_new_sample_spacing(by_slice=by_slice, new_spacing=new_spacing)]

            # Check that new spacing is now a nested list.
            if not all(isinstance(ii, list) for ii in new_spacing):
                raise TypeError(f"THe new_spacing variable should now be represented as a nested list.")

        # Set spacing for resampling. Note that new_spacing should now either be None or a nested list, i.e. a list
        # containing other lists.
        self.new_spacing: None | list[list[float | None]] = new_spacing

        # Set anti-aliasing.
        self.anti_aliasing: bool = anti_aliasing

        # Check that smoothing beta lies between 0.0 and 1.0.
        if anti_aliasing:
            if smoothing_beta <= 0.0 or smoothing_beta > 1.0:
                raise ValueError(
                    f"The value of the smoothing_beta parameter should lie between 0.0 and 1.0, "
                    f"not including 0.0. Found: {smoothing_beta}")

        # Set smoothing beta.
        self.smoothing_beta: float = smoothing_beta

    @staticmethod
    def _check_new_sample_spacing(by_slice, new_spacing):
        # Checks whether sample spacing is correctly set, and parses it.

        # Parse value to list of floating point values to facilitate checks.
        if isinstance(new_spacing, (int, float)):
            new_spacing = [new_spacing]

        # Convert to floating point values.
        new_spacing: list[float | None] = [float(new_spacing_element) for new_spacing_element in new_spacing]

        if by_slice:
            # New spacing is expect to consist of at most two values when the experiment is based on slices. A
            # placeholder for the z-direction is set here.
            if len(new_spacing) == 1:
                # This creates isotropic spacing.
                new_spacing = [None, new_spacing[0], new_spacing[0]]

            elif len(new_spacing) == 2:
                # Insert a placeholder for the z-direction.
                new_spacing.insert(0, None)

            else:
                raise ValueError(
                    f"The desired voxel spacing for in-slice resampling should consist of two "
                    f"elements. Found: {len(new_spacing)} elements.")
        else:
            if len(new_spacing) == 1:
                # This creates isotropic spacing.
                new_spacing = [new_spacing[0], new_spacing[0], new_spacing[0]]

            elif len(new_spacing) == 3:
                # Do nothing.
                pass

            else:
                raise ValueError(
                    f"The desired voxel spacing for volumetric resampling should consist of three "
                    f"elements. Found: {len(new_spacing)} elements.")

        return new_spacing


def get_image_interpolation_settings() -> list[dict[str, Any]]:
    return [
        setting_def("new_spacing", "float", to_list=True, test=[1.0, 1.0, 1.0]),
        setting_def("spline_order", "int", test=2),
        setting_def("anti_aliasing", "bool", test=False),
        setting_def("smoothing_beta", "float", test=0.75)
    ]


@dataclass
class MaskInterpolationSettingsClass:
    """
    Parameters related to mask interpolation / resampling. MIRP registers the mask to an interpolated image based,
    and fewer parameters can be set compared to image interpolation / resampling (
    :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass`).

    Parameters
    ----------
    roi_spline_order: int, optional, default: 1
        Sets the spline order used for spline interpolation. mirp uses `scipy.ndimage.map_coordinates
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.map_coordinates.html#scipy.ndimage
        .map_coordinates>`_ internally. Spline orders 0, 1, and 3 refer to nearest neighbour, linear interpolation
        and cubic interpolation, respectively.

    roi_interpolation_mask_inclusion_threshold: float, optional, default: 0.5
        Threshold for partially masked voxels after interpolation. All voxels with a value equal to or greater than
        this threshold are assigned to the mask.

    **kwargs: dict, optional
        Unused keyword arguments.
    """

    def __init__(
            self,
            roi_spline_order: int = 1,
            roi_interpolation_mask_inclusion_threshold: float = 0.5,
            **kwargs):

        # Check if the spline order is valid.
        if roi_spline_order < 0 or roi_spline_order > 5:
            raise ValueError(
                f"The interpolation spline order for the ROI should be an integer between 0 and 5. Found:"
                f" {roi_spline_order}")

        # Set spline order.
        self.spline_order = roi_spline_order

        # Check if the inclusion threshold is between 0 and 1.
        if roi_interpolation_mask_inclusion_threshold <= 0.0 or roi_interpolation_mask_inclusion_threshold > 1.0:
            raise ValueError(
                f"The inclusion threshold for the ROI mask should be between 0.0 and 1.0, excluding 0.0. "
                f"Found: {roi_interpolation_mask_inclusion_threshold}")

        self.incl_threshold = roi_interpolation_mask_inclusion_threshold


def get_mask_interpolation_settings() -> list[dict[str, Any]]:
    return [
        setting_def("roi_spline_order", "int", xml_key="spline_order", class_key="spline_order", test=2),
        setting_def("roi_interpolation_mask_inclusion_threshold", "float", xml_key="incl_threshold",
                    class_key="incl_threshold", test=0.25)
    ]



from typing import Generator, Iterable, Any
import copy, logging, sys

from mirp._data_import.generic_file import ImageFile
from mirp.settings.generic import SettingsClass
from mirp.utilities.parallel import parse_parallel_backend, start_parallel_cluster, cluster_exists, shutdown_cluster, message_parallel_process
from mirp.utilities.parallel_ray import ray_remote, ray_get
from mirp._workflows.standardWorkflow import StandardWorkflow


def deep_learning_preprocessing(
        output_slices: bool = False,
        crop_size: None | list[float] | list[int] = None,
        image_export_format: str = "dict",
        write_file_format: str = "numpy",
        export_images: None | bool = None,
        write_images: None | bool = None,
        write_dir: None | str = None,
        num_cpus: None | int = None,
        parallel_backend: None | str = None,
        **kwargs
) -> None | list[Any]:
    """
    Pre-processes images for deep learning.

    Parameters
    ----------
    output_slices: bool, optional, default: False
        Determines whether separate slices should be extracted.

    crop_size: list of float or list of int, optional, default: None
        Size to which the images and masks should be cropped. Images and masks are cropped around the center of the
        mask(s).

        .. note::
            MIRP follows the numpy convention for indexing (*z*, *y*, *x*). The final element always corresponds to the
            *x* dimension.

    image_export_format: {"dict", "native", "numpy"}, default: "dict"
        Return format for processed images and masks. ``"dict"`` returns dictionaries of images and masks as numpy
        arrays and associated characteristics. ``"native"`` returns images and masks in their internal format.
        ``"numpy"`` returns images and masks in numpy format. This argument is only used if ``export_images=True``.

    write_file_format: {"nifti", "numpy"}, default: "numpy"
        File format for processed images and masks. ``"nifti"`` writes images and masks in the NIfTI file format,
        and ``"numpy"`` writes images and masks as numpy files. This argument is only used if ``write_images=True``.

    export_images: bool, optional
        Determines whether processed images and masks should be returned by the function.

    write_images: bool, optional
        Determines whether processed images and masks should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    write_dir: str, optional
        Path to directory where processed images and masks should be written. If not set, processed images and masks
        are returned by this function. Required if ``write_images=True``.

    num_cpus: int, optional, default: None
        Number of CPU nodes that should be used for parallel processing. Image and mask processing can be
        parallelized using the ``ray`` or ``joblib`` packages. If a ray cluster is defined by the user, this cluster
        will be used instead. By default, image and mask processing are processed sequentially.

    parallel_backend: {"none", "ray", "joblib"}, optional, default: "none"
        Type of backend to use. Default is the sequential backend (``"none"``). Alternative backends are ``"ray"`` and
        ``"joblib"``, which rely on the ray and joblib libraries respectively.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`~mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings (notably
        :class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`,
        :class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`), among others.

    Returns
    -------
    None | list[Any]
        List of images and masks in the format indicated by ``image_export_format``, if ``export_images=True``.

    See Also
    --------
    Keyword arguments can be provided to configure the following:

    * image and mask import (:func:`~mirp.data_import.import_image_and_mask.import_image_and_mask`)
    * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
    * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
    * image interpolation / resampling (:class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
      :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
    * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)

    """

    # Configure logger
    logging.basicConfig(
        format="%(levelname)s\t: %(processName)s \t %(asctime)s \t %(message)s",
        level=logging.INFO,
        stream=sys.stdout
    )

    backend = parse_parallel_backend(backend=parallel_backend, num_cpus=num_cpus)
    external_cluster = cluster_exists(backend=backend)
    start_parallel_cluster(backend=backend, num_cpus=num_cpus)

    # Switch to sequential backend if ray cluster is not formed somehow.
    if backend == "ray" and not cluster_exists(backend=backend):
        backend = "none"

    logging.info(message_parallel_process(backend=backend, num_cpus=num_cpus))

    if backend == "none":
        workflows = list(_base_deep_learning_preprocessing(
            export_images=export_images,
            write_images=write_images,
            write_dir=write_dir,
            **kwargs)
        )

        results = [
            workflow.deep_learning_conversion(
                output_slices=output_slices,
                crop_size=crop_size,
                image_export_format=image_export_format,
                write_file_format=write_file_format
            )
            for workflow in workflows
        ]

    elif backend == "ray":
        results = [
            _ray_extractor.remote(
                workflow=workflow,
                output_slices=output_slices,
                crop_size=crop_size,
                image_export_format=image_export_format,
                write_file_format=write_file_format
            )
            for workflow in _base_deep_learning_preprocessing(
                export_images=export_images,
                write_images=write_images,
                write_dir=write_dir,
                **kwargs
            )
        ]
        results = ray_get(results)

    elif backend == "joblib":
        from joblib import Parallel, delayed
        results = Parallel(n_jobs=num_cpus)(
            delayed(workflow.deep_learning_conversion)(
                output_slices=output_slices,
                crop_size=crop_size,
                image_export_format=image_export_format,
                write_file_format=write_file_format
            ) for workflow in _base_deep_learning_preprocessing(
                export_images=export_images,
                write_images=write_images,
                write_dir=write_dir,
                **kwargs
            )
        )

    else:
        raise ValueError(f"parallel_backend is expected to be one of 'none', 'ray' or 'joblib'. Found: {backend}")

    if not external_cluster:
        shutdown_cluster(backend=backend)

    return results


@ray_remote
def _ray_extractor(
        workflow: StandardWorkflow,
        output_slices: bool = False,
        crop_size: None | list[float] | list[int] = None,
        image_export_format: str = "numpy",
        write_file_format: str = "numpy"
):
    # Limit internal threading by third-party libraries.
    from mirp.utilities.parallel_ray import limit_inner_threads
    limit_inner_threads()

    return workflow.deep_learning_conversion(
        output_slices=output_slices,
        crop_size=crop_size,
        image_export_format=image_export_format,
        write_file_format=write_file_format
    )


def deep_learning_preprocessing_generator(
        output_slices: bool = False,
        crop_size: None | list[float] | list[int] = None,
        image_export_format: str = "dict",
        write_file_format: str = "numpy",
        export_images: None | bool = None,
        write_images: None | bool = None,
        write_dir: None | str = None,
        num_cpus: None | int = None,
        parallel_backend: None | str = None,
        **kwargs
) -> Generator[Any, None, None]:
    """
    Generator for pre-processing images for deep learning.

    Parameters
    ----------
    output_slices: bool, optional, default: False
        Determines whether separate slices should be extracted.

    crop_size: list of float or list of int, optional, default: None
        Size to which the images and masks should be cropped. Images and masks are cropped around the center of the
        mask(s).

        .. note::
            MIRP follows the numpy convention for indexing (*z*, *y*, *x*). The final element always corresponds to the
            *x* dimension.

    image_export_format: {"dict", "native", "numpy"}, default: "dict"
        Return format for processed images and masks. ``"dict"`` returns dictionaries of images and masks as numpy
        arrays and associated characteristics. ``"native"`` returns images and masks in their internal format.
        ``"numpy"`` returns images and masks in numpy format. This argument is only used if ``export_images=True``.

    write_file_format: {"nifti", "numpy"}, default: "numpy"
        File format for processed images and masks. ``"nifti"`` writes images and masks in the NIfTI file format,
        and ``"numpy"`` writes images and masks as numpy files. This argument is only used if ``write_images=True``.

    export_images: bool, optional
        Determines whether processed images and masks should be returned by the function.

    write_images: bool, optional
        Determines whether processed images and masks should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    write_dir: str, optional
        Path to directory where processed images and masks should be written. If not set, processed images and masks
        are returned by this function. Required if ``write_images=True``.

    num_cpus: int, optional, default: None
        Number of CPU nodes that should be used for parallel processing. Image and mask processing can be
        parallelized using the ``joblib`` package. By default, image and mask processing are processed sequentially.

    parallel_backend: {"none", "joblib"}, optional, default: "none"
        Type of backend to use. Default is the sequential backend (``"none"``). ``"joblib"`` can be used as
        an alternative backend. ``"ray"`` cannot be used in a generator context, because only a single worker will be
        used.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`~mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings (notably
        :class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`,
        :class:`~mirp.settings.settingsPerturbation.ImagePerturbationSettingsClass`), among others.

    Yields
    -------
    None | list[Any]
        List of images and masks in the format indicated by ``image_export_format``, if ``export_images=True``.

    See Also
    --------
    Keyword arguments can be provided to configure the following:

    * image and mask import (:func:`~mirp.data_import.import_image_and_mask.import_image_and_mask`)
    * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
    * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
    * image interpolation / resampling (:class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
      :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
    * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)

    """

    # Configure logger
    logging.basicConfig(
        format="%(levelname)s\t: %(processName)s \t %(asctime)s \t %(message)s",
        level=logging.INFO,
        stream=sys.stdout
    )

    # Do not allow ray as a backend.
    backend = parse_parallel_backend(backend=parallel_backend, num_cpus=num_cpus, ray_allowed=False)
    external_cluster = cluster_exists(backend=backend)
    start_parallel_cluster(backend=backend, num_cpus=num_cpus)

    # Switch to sequential backend if ray cluster is not formed somehow.
    if backend == "ray" and not cluster_exists(backend=backend):
        backend = "none"

    logging.info(message_parallel_process(backend=backend, num_cpus=num_cpus))

    if backend == "none":
        for workflow in _base_deep_learning_preprocessing(
                export_images=export_images,
                write_images=write_images,
                write_dir=write_dir,
                **kwargs
        ):
            yield workflow.deep_learning_conversion(
                output_slices=output_slices,
                crop_size=crop_size,
                image_export_format=image_export_format,
                write_file_format=write_file_format
            )

    elif backend == "joblib":
        from joblib import Parallel, delayed
        parallel_gen = Parallel(n_jobs=num_cpus, return_as="generator")(
            delayed(workflow.deep_learning_conversion)(
                output_slices=output_slices,
                crop_size=crop_size,
                image_export_format=image_export_format,
                write_file_format=write_file_format
            ) for workflow in _base_deep_learning_preprocessing(
                export_images=export_images,
                write_images=write_images,
                write_dir=write_dir,
                **kwargs
            )
        )

        yield from parallel_gen

    else:
        raise ValueError(f"parallel_backend is expected to be one of 'none' or 'joblib'. Found: {backend}")

    if not external_cluster:
        shutdown_cluster(backend=backend)


def _base_deep_learning_preprocessing(
        image,
        mask=None,
        sample_name: None | str | list[str] = None,
        image_name: None | str | list[str] = None,
        image_file_type: None | str = None,
        image_modality: None | str | list[str] = None,
        image_sub_folder: None | str = None,
        mask_name: None | str | list[str] = None,
        mask_file_type: None | str = None,
        mask_modality: None | str | list[str] = None,
        mask_sub_folder: None | str = None,
        roi_name: None | str | list[str] | dict[str, str] = None,
        association_strategy: None | str | list[str] = None,
        settings: None | str | SettingsClass | list[SettingsClass] = None,
        stack_masks: str = "auto",
        stack_images: str = "auto",
        write_images: None | bool = None,
        export_images: None | bool = None,
        write_dir: None | str = None,
        **kwargs
):
    from mirp.data_import.import_image_and_mask import import_image_and_mask
    from mirp.settings.import_config_parameters import import_configuration_settings

    # Infer write_images, export_images based on write_dir.
    if write_images is None:
        write_images = write_dir is not None
    if export_images is None:
        export_images = write_dir is None

    if not write_images:
        write_dir = None

    if write_images and write_dir is None:
        raise ValueError("write_dir argument should be provided for writing images and masks to.")

    if not write_images and not export_images:
        raise ValueError(f"write_images and export_images arguments cannot both be False.")

    # Import settings (to provide immediate feedback if something is amiss).
    if isinstance(settings, str):
        settings = import_configuration_settings(
            compute_features=False,
            path=settings
        )
    elif isinstance(settings, SettingsClass):
        settings = [settings]
    elif isinstance(settings, Iterable) and all(isinstance(x, SettingsClass) for x in settings):
        settings = list(settings)
    elif settings is None:
        settings = import_configuration_settings(
            compute_features=False,
            **kwargs
        )
    else:
        raise TypeError(
            f"The 'settings' argument is expected to be a path to a configuration xml file, "
            f"a SettingsClass object, or a list thereof. Found: {type(settings)}."
        )

    image_list = import_image_and_mask(
        image=image,
        mask=mask,
        sample_name=sample_name,
        image_name=image_name,
        image_file_type=image_file_type,
        image_modality=image_modality,
        image_sub_folder=image_sub_folder,
        mask_name=mask_name,
        mask_file_type=mask_file_type,
        mask_modality=mask_modality,
        mask_sub_folder=mask_sub_folder,
        roi_name=roi_name,
        association_strategy=association_strategy,
        stack_images=stack_images,
        stack_masks=stack_masks
    )

    yield from _generate_dl_preprocessing_workflows(
        image_list=image_list,
        settings=settings,
        write_dir=write_dir,
        write_images=write_images,
        export_images=export_images
    )


def _generate_dl_preprocessing_workflows(
        image_list: list[ImageFile],
        settings: list[SettingsClass],
        write_dir: None | str,
        write_images: bool,
        export_images: bool
) -> Generator[StandardWorkflow, None, None]:

    for image_file in image_list:
        for current_settings in settings:

            # Update settings to remove settings that may cause problems.
            current_settings.feature_extr.families = "none"
            current_settings.img_transform.feature_settings.families = "none"
            current_settings.perturbation.crop_around_roi = False
            current_settings.roi_resegment.resegmentation_method = "none"

            if current_settings.perturbation.noise_repetitions is None or \
                    current_settings.perturbation.noise_repetitions == 0:
                noise_repetition_ids = [None]
            else:
                noise_repetition_ids = list(range(current_settings.perturbation.noise_repetitions))

            if current_settings.perturbation.rotation_angles is None or len(
                    current_settings.perturbation.rotation_angles) == 0 or all(
                x == 0.0 for x in current_settings.perturbation.rotation_angles
            ):
                rotation_angles = [None]
            else:
                rotation_angles = copy.deepcopy(current_settings.perturbation.rotation_angles)

            if current_settings.perturbation.translation_fraction is None or len(
                current_settings.perturbation.translation_fraction) == 0 or all(
                x == 0.0 for x in current_settings.perturbation.translation_fraction
            ):
                translations = [None]
            else:
                config_translation = copy.deepcopy(current_settings.perturbation.translation_fraction)
                translations = []
                for translation_x in config_translation:
                    for translation_y in config_translation:
                        if not current_settings.general.by_slice:
                            for translation_z in config_translation:
                                translations += [(translation_z, translation_y, translation_x)]
                        else:
                            translations += [(0.0, translation_y, translation_x)]

            if current_settings.img_interpolate.new_spacing is None or len(
                    current_settings.img_interpolate.new_spacing) == 0 or all(
                x == 0.0 for x in current_settings.img_interpolate.new_spacing
            ):
                spacings = [None]
            else:
                spacings = copy.deepcopy(current_settings.img_interpolate.new_spacing)

            for noise_repetition_id in noise_repetition_ids:
                for rotation_angle in rotation_angles:
                    for translation in translations:
                        for spacing in spacings:
                            yield StandardWorkflow(
                                image_file=copy.deepcopy(image_file),
                                write_dir=write_dir,
                                settings=current_settings,
                                settings_name=current_settings.general.config_str,
                                write_features=False,
                                export_features=False,
                                write_images=write_images,
                                export_images=export_images,
                                noise_iteration_id=noise_repetition_id,
                                rotation=rotation_angle,
                                translation=translation,
                                new_image_spacing=spacing
                            )
